import numpy as np
import shapely.geometry
from shapely import affinity
from scipy.integrate import quad
import matplotlib.pyplot as plt

def simulate_line_measurements( angs, strain, ytrans, ztrans, sample_corners, show_geom=False ):
    '''Simulate parallel ray integrals over a strain field in a 3D sample by looping 
       over z-positions. Convention here is that coordinate x is along the xray for an
       angle of zero degrees. The sample is represented as a stack of 2D polygons.

    Input:
        angs            : array of angular settings in degrees of the sample
        strain          : function that takes a coordinate array x=(x,y,z) and returns a numpy (3x3) strain tensor 
        ytrans          : sample y coordinate translation for measurements
        ztrans          : sample z coordinate translation for measurements
        sample_corners  : List containing the sample corners for each ztrans
        show_geom       : debug option, plots the sample and xray for each measurement

    Returns:
        Y      : is N by 1 array of the measurement values
        sig_m  : is N by 1 array of the standard deviations of the measurement values
        entry  : is a 3 by N array of the coordinates where the ray enters the sample (each col is x, y, z)
        exit   : is a 3 by N array of the coordinates where the ray leaves the sample (each col is x, y,z)
        nhat   : is a 3 by N array of directions of each ray (each col is a unit vector)
        kappa  : is a 3 by N array that defines the direction of strain measured (each col is a unit vector)
        L      : is a 1 by N array of the total irradiated lengths of each ray (ie distance between entry and exit)
        nsegs  : is a 1 by N array of all ones (number of segments for each ray) this is useful when we go to non convex geometry

        (If we have non convex geometry then
        nsegs is a 1 by N array containing the number of times that the ray enters and exits the sample
        entry becomes a p*3 by N array where p is max(nsegs) and each col is teh stacked entries so [first entry x, first entry y, first entry z, second entry x, second entry y, and so on]
        exit becomes a p*3 by N array with a similar story)

    NOTE: If a ray is outside the sample all values in the output arrays are put to zero for this 
          measurement setting. The same is true if the ray only graces a corner point.
    '''

    # Instantiate lists to contain all measurements
    all_Y, all_sig_m, all_entry, all_exit, all_nhat, all_kappa, all_L, all_nsegs = [],[],[],[],[],[],[],[]
    xray_endpoints = get_xray_endpoints( sample_corners )

    # Loop over all experimental settings
    for ang in angs:

        # select a random kappa for this angle
        strain_dir = np.random.rand( 3 )
        strain_dir = strain_dir/np.linalg.norm(strain_dir)

        for z,sample_bounds in zip(ztrans, sample_corners):

            # extract the sample geometry at this z-coordinate
            sample_polygon = shapely.geometry.Polygon(sample_bounds)

            for y in ytrans:

                # Translate and rotate the xray endpoints according to ytrans and angle
                c, s = np.cos( np.radians( -ang ) ), np.sin( np.radians( -ang ) )
                rotz = np.array([[c,-s],[s,c]])
                rx = rotz.dot( xray_endpoints + np.array([[0,0],[-y,-y]]) )
                xray_polygon = shapely.geometry.LineString( [ rx[:,0], rx[:,1] ] )
                
                if show_geom:
                    # Show sample and xray polygons
                    xc,yc = xray_polygon.xy
                    plt.plot(xc,yc)
                    xc,yc = sample_polygon.exterior.xy
                    plt.plot(xc,yc)
                    plt.title(' angle='+str(ang)+r'$^o$,   y_translation='+str(y)+',   z_translation='+str(z))
                    plt.axis('equal')
                    plt.show()

                # compute the intersections between beam and sample
                intersection_points = get_intersection( xray_polygon, sample_polygon, z  )
                
                if intersection_points is None:
                    # If a measurement missed the sample or graced a corner, we skipp ahead
                    continue
                else:
                    # make a measurement at the current setting
                    Y,sig_m,entry,exit,nhat,kappa,L,nsegs = measure( strain, intersection_points, strain_dir )
                
                    # save the measurement results in global lists
                    all_Y.append( Y )
                    all_sig_m.append( sig_m )
                    all_entry.append( entry )
                    all_exit.append( exit )
                    all_nhat.append( nhat )
                    all_kappa.append( kappa )
                    all_L.append( L )
                    all_nsegs.append( nsegs )

    # repack lists of measurements into numpy arrays of desired format
    Y, sig_m, entry, exit, nhat, kappa, L, nsegs = repack(all_Y, all_sig_m, all_entry, all_exit, all_nhat, all_kappa, all_L, all_nsegs)

    return Y, sig_m, entry, exit, nhat, kappa, L, nsegs


def get_xray_endpoints( sample_xy_corners ):
    '''Calculate endpoitns of xray line segement. The lenght of the 
    line segment is adapted to make sure xray always convers the full
    length of the sample.
    '''
    xmin = np.min( [ np.min( [c[0] for c in sc] ) for sc in sample_xy_corners ] )
    xmax = np.max( [ np.max( [c[0] for c in sc] ) for sc in sample_xy_corners ] )
    ymin = np.min( [ np.min( [c[1] for c in sc] ) for sc in sample_xy_corners ] )
    ymax = np.max( [ np.max( [c[1] for c in sc] ) for sc in sample_xy_corners ] )
    D = np.sqrt( (xmax-xmin)**2 + (ymax-ymin)**2 )
    return np.array([ [-1.1*D, 1.1*D], [0,0] ])                      


def get_intersection(xray_polygon, sample_polygon, z ):
    '''Compute the 3d coordinates of intersection between xray and
    sample.
    '''
    intersection = sample_polygon.intersection( xray_polygon )
    if intersection.is_empty:
        # we missed the sample with the beam
        intersection_points = None
    elif isinstance(intersection, shapely.geometry.linestring.LineString):
        # we got a single line segment intersection
        intersection_points = np.zeros( (2,3) )
        intersection_points[:2,:2] = np.array( intersection.xy ).T
        intersection_points[:,2] = -z
    elif isinstance(intersection, shapely.geometry.multilinestring.MultiLineString):
        # we got multiple line segments intersection
        intersection_points = np.zeros( (2*len(intersection.geoms),3) )
        for i,line_segment in enumerate(intersection.geoms):
            intersection_points[2*i:2*(i+1),:2] = np.array( line_segment.xy ).T
        intersection_points[:,2] = -z
    
    return intersection_points


def repack(all_Y, all_sig_m, all_entry, all_exit, all_nhat, all_kappa, all_L, all_nsegs):
    '''Repack global measurement list into numpy arrays of desired format.
    '''

    N = len(all_Y)

    p = max( max(all_nsegs), 1 )
    nsegs = np.array(all_nsegs).reshape( 1,N )
    Y = np.array(all_Y).reshape( N, )
    sig_m  = np.array(all_sig_m).reshape( N,1 )
    L = np.array(all_L).reshape( 1,N )

    entry = np.zeros( (3*p, N) )
    for i,en in enumerate(all_entry):
        entry[:len(en[:]),i] = en[:]

    exit = np.zeros( (3*p, N) )
    for i,ex in enumerate(all_exit):
        exit[:len(ex[:]),i] = ex[:]

    nhat = np.array( all_nhat ).T
    kappa = np.array( all_kappa ).T
    return Y, sig_m, entry, exit, nhat, kappa, L, nsegs


def measure( strain, intersection_points, strain_dir ):
    '''Perform a single measurement integrating by quadrature over
    the strain function.
    '''

    nsegs = intersection_points.shape[0]//2

    kappa = list(strain_dir)
    entry,exit = [],[]
    Y = 0
    L = 0

    p1 = intersection_points[0,:] 
    p2 = intersection_points[1,:]
    
    length = np.linalg.norm(p2-p1)
    xray_direction = (p2-p1)/length
    nhat = list(xray_direction)
    theta = np.arccos( -xray_direction.dot(strain_dir) )
    sig_m = (np.pi - theta)*(1e-4/np.pi) + 1e-4 # higher scattering angles are more certain measurements

    for i in range( nsegs ):
        p1 = intersection_points[2*i,:] 
        p2 = intersection_points[2*i+1,:]
        entry.extend( list(p1) )
        exit.extend( list(p2) )
        length = np.linalg.norm(p2-p1)
        def func(s): 
            epsilon = strain(p1 + s*xray_direction)
            return strain_dir.dot( epsilon ).dot( strain_dir.T )
        Y += quad(func, 0, length)[0]
        L += length
    Y = Y/L
        
    return Y, sig_m, entry, exit, nhat, kappa, L, nsegs



import numpy as np
import copy
import matplotlib.pyplot as plt

def match_and_label( pixelated_grains, id11_grains, angthres=3.0 ):
    """Segment a stack of 2d grain slices into 3d grains by orientation and z-overlap.

    This function iterates over stacks for pixelated grain masks and corresponding Id11
    grain objects to create a numpy 3d volume where each voxel has a uniqe integer number
    corresponding to the grain index. Void regions are set to 0. The index at a voxel
    corresponds to a Id11 grain object in an output dictionary of Id11 grains.

    Grains are considered to be one and the same if and only if:
        #1: They have at least one overlaping voxel in z.
        #2: They have orientation matrices resutling in a smaller angular deviation (dgrs)
            than angthres. This number represents the sum of angular devation at each
            orientation vector formed by taking the arccos of the three scalar products.
            
    Args:
        pixelated_grains (:obj:`list` of :obj:`lists` of :obj:`numpy arrays`): Pixelated masks of the grains
        id11_grains (:obj:`list` of :obj:`lists` of :obj:`ImageD11 Grain`): ImageD11 grain representations
            corresponding to the masks in pixelated_grains.
        angthres (:obj:`float`): Criteria threshold for mapping grains across z. Defaults to 3.0.
    
    Returns:
        sample (:obj:`numpy array`): Voxelated representation of full sample. the value at sample[i,j,k]
        is ```0``` if void and otherwise equal to the index of the grain in ```grains``` 
        grains (:obj:`dict` of :obj:`dict` of :obj:`ImageD11 Grain`): ImageD11 grain representations
            corresponding to ```sample``` index values such that grains["3"]["7"] gives the 7th slice if grain
            with index 3 in ```sample```.

    """

    grain_indx = 1
    grains = {}
    dx,dy = pixelated_grains[0][0].shape
    sample = np.zeros( (dx, dy, len(pixelated_grains) ), dtype=int )

    for j in range(len(pixelated_grains[0])):
        mask = pixelated_grains[0][j]
        sample[mask,0] = grain_indx
        grains[str(grain_indx)] = { str(0) : id11_grains[0][j] }
        grain_indx += 1

    for i in range(1, sample.shape[2]):
        for j in range(len(pixelated_grains[i])):
            found_match = False
            mask = pixelated_grains[i][j]
            overlap = ( sample[:,:,i-1] * mask ) > 0
            if np.sum( overlap ) > 0:
                indices = np.unique(sample[overlap,i-1])
                for indx in indices[indices!=0]:
                    u1 = grains[str(indx)][str(i-1)].u
                    u2 = id11_grains[i][j].u
                    angdiff = np.sum( [ np.degrees( np.arccos(u1[:,k].dot(u2[:,k])) ) for k in range(3) ] )
                    if angdiff < angthres:
                        sample[mask,i] = indx
                        found_match = True
                        grains[str(indx)][str(i)] = id11_grains[i][j]
                        break

            if not found_match:
                sample[mask,i] = grain_indx
                grains[str(grain_indx)] = { str(i) : id11_grains[i][j] }
                grain_indx += 1
            
    return sample, grains







import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import Polygon as shapelyPolygon
from shapely.geometry import Point as shapelyPoint

def mesh_from_polygon( w, poly ):
    """Generate a quadratile 2d mesh over a Shapely polygon object.

    Args:
        w (:obj:`float`): Element side length.
        poly (:obj:`Shapely Polygon`): Polygon object to be meshed. The output mesh 
            follows the :obj:`poly` coordinate system.
    
    Returns:
        (:obj:`numpy array`) The nodal coordinates of each element in the mesh.

    """
    x = poly.exterior.xy[0]
    y = poly.exterior.xy[1]
    xmin,xmax = np.min(x), np.max(x)
    ymin,ymax = np.min(y), np.max(y)

    ll = [xmin, ymin]
    mesh = []
    while(ll[1]<=ymax-w):
        if poly.contains( shapelyPoint([(ll[0]+w/2., ll[1]+w/2.)]) ):
            mesh.append([ll[0], ll[1], ll[0]+w, ll[1], ll[0]+w, ll[1]+w, ll[0], ll[1]+w])
        if ll[0] < xmax - w:
            ll[0] = ll[0] + w
        else:
            ll[0] = xmin
            ll[1] = ll[1] + w
    return np.array(mesh)

def mesh_to_pixels( mesh, element_width, shape, values ):
    """Generate pixelated map from a mesh coordinate matrix.

    Args:
        mesh (:obj:`numpy array`): Element nodal coordinates
        element_width (:obj:`float`): Element side length.
        shape  (:obj:`tuple`): Output array shape
        values (:obj:`numpy array`): Per elemetn values to fill the output array with.

    Returns:
        (:obj:`numpy array`) Pixelated map of input field stored in values.

    """
    mask = np.full(shape, np.nan)
    for i,(xc,yc) in enumerate(zip( np.mean(mesh[:,0::2],axis=1), np.mean(mesh[:,1::2],axis=1) )):
        xindx =  -np.ceil( -xc/element_width ).astype(int) + shape[0]//2
        yindx =   np.ceil( yc/element_width ).astype(int) + shape[1]//2
        mask[yindx,xindx] = values[i]
    return mask

def show_mesh( mesh, poly ):
    print(mesh.shape)
    x = poly.exterior.xy[0]
    y = poly.exterior.xy[1]
    for e in mesh:
        for i in [0,2,4]:
            plt.plot([e[i],e[i+2]],[e[i+1],e[i+3]],'-ko')
        plt.plot([e[-2],e[0]],[e[-1],e[1]],'-ko')
    for i in range(len(x)-1):
        plt.plot([x[i],x[i+1]],[y[i],y[i+1]],'k')
    plt.plot([x[-1],x[0]],[y[-1],y[0]],'k')
    plt.grid(True)
    plt.show()



from __future__ import print_function

""" Modified from ImageD11 transform.py by Axel Henningsson 2021
    This is only good for the scanning_3DXRD scenario when the t_y==0 and 
    t_z==0 for all reflections, the idea is to allow for per measurement
    grain cms positions to be passed via t_x,t_y,t_z params."""

import numpy as np
from ImageD11 import gv_general
from numpy import radians, degrees

def cross_product_2x2(a, b):
    """ returns axb for two len(3) vectors a,b"""
    assert len(a) == len(b) == 3
    return np.array([a[1] * b[2] - a[2] * b[1],
                     a[2] * b[0] - a[0] * b[2],
                     a[0] * b[1] - a[1] * b[0]])

def detector_rotation_matrix(tilt_x, tilt_y, tilt_z):
    """
    Return the tilt matrix to apply to peaks
    tilts in radians
    typically applied to peaks rotating around beam center
    """
    r1 = np.array([[np.cos(tilt_z), -np.sin(tilt_z), 0],  # note this is r.h.
                   [np.sin(tilt_z), np.cos(tilt_z), 0],
                   [0,    0, 1]], np.float)
    r2 = np.array([[np.cos(tilt_y), 0, np.sin(tilt_y)],
                   [0, 1,   0],
                   [-np.sin(tilt_y), 0, np.cos(tilt_y)]], np.float)
    r3 = np.array([[1,          0,       0],
                   [0,  np.cos(tilt_x), -np.sin(tilt_x)],
                   [0,  np.sin(tilt_x), np.cos(tilt_x)]], np.float)
    r2r1 = np.dot(np.dot(r3, r2), r1)
    return r2r1

def compute_xyz_lab(peaks,
                    y_center=0., y_size=0., tilt_y=0.,
                    z_center=0., z_size=0., tilt_z=0.,
                    tilt_x=0.,
                    distance=0.,
                    # detector_orientation=((1,0),(0,1)),
                    o11=1.0, o12=0.0, o21=0.0, o22=-1.0,
                    **kwds):
    """
    Peaks is a 2 d array of x,y
    yc is the centre in y
    ys is the y pixel size
    ty is the tilt around y
    zc is the centre in z
    zs is the z pixel size
    tz is the tilt around z
    dist is the sample - detector distance
    detector_orientation is a matrix to apply to peaks arg to get
    ImageD11 convention
         (( 0, 1),( 1, 0)) for ( y, x)
         ((-1, 0),( 0, 1)) for (-x, y)
         (( 0,-1),(-1, 0)) for (-y,-x)
      etc...
    """
    assert len(peaks) == 2, "peaks must be a 2D array"
    # Matrix for the tilt rotations
    r2r1 = detector_rotation_matrix(tilt_x, tilt_y, tilt_z)
    # Peak positions in 3D space
    #  - apply detector orientation
    peaks_on_detector = np.array(peaks)
    peaks_on_detector[0, :] = (peaks_on_detector[0, :] - z_center) * z_size
    peaks_on_detector[1, :] = (peaks_on_detector[1, :] - y_center) * y_size
    #
    detector_orientation = [[o11, o12], [o21, o22]]
    # logging.debug("detector_orientation = "+str(detector_orientation))
    flipped = np.dot(np.array(detector_orientation, np.float),
                     peaks_on_detector)
    #
    vec = np.array([np.zeros(flipped.shape[1]),  # place detector at zero,
                    # sample at -dist
                    flipped[1, :],             # x in search, frelon +z
                    flipped[0, :]], np.float)     # y in search, frelon -y
    # Position of diffraction spots in 3d space after detector tilts about
    # the beam centre on the detector
    rotvec = np.dot(r2r1, vec)
    # Now add the distance (along x)
    rotvec[0, :] = rotvec[0, :] + distance
    return rotvec


def compute_tth_eta(peaks,
                    y_center=0., y_size=0., tilt_y=0.,
                    z_center=0., z_size=0., tilt_z=0.,
                    tilt_x=0.,
                    distance=0.,
                    # detector_orientation=((1,0),(0,1)),
                    o11=1.0, o12=0.0, o21=0.0, o22=-1.0,
                    t_x=0.0, t_y=0.0, t_z=0.0,
                    omega=None,  # == phi at chi=90
                    wedge=0.0,  # Wedge == theta on 4circ
                    chi=0.0,  # == chi - 90
                    **kwds):  # spare args are ignored
    """
    0/10 for style
    """

    peaks_xyz = compute_xyz_lab(
        peaks,
        y_center=y_center, y_size=y_size, tilt_y=tilt_y,
        z_center=z_center, z_size=z_size, tilt_z=tilt_z,
        tilt_x=tilt_x,
        distance=distance,
        # detector_orientation=((1,0),(0,1)),
        o11=o11, o12=o12, o21=o21, o22=o22)

    tth, eta = compute_tth_eta_from_xyz(
        peaks_xyz,
        t_x=t_x, t_y=t_y, t_z=t_z,
        omega=omega,
        wedge=wedge,
        chi=chi)

    return tth, eta


def compute_tth_eta_from_xyz(peaks_xyz, omega,
                             t_x=0.0, t_y=0.0, t_z=0.0,
                             #       == phi at chi=90
                             wedge=0.0,  # Wedge == theta on 4circ
                             chi=0.0,  # == chi - 90
                             **kwds):  # last line is for laziness -
    """
    Peaks is a 3 d array of x,y,z peak co-ordinates
    crystal_translation is the position of the grain giving rise
    to a diffraction spot
                 in x,y,z ImageD11 co-ordinates
                 x,y with respect to axis of rotation and or beam centre ??
                 z with respect to beam height, z centre
    omega data needed if crystal translations used
    """
    assert len(peaks_xyz) == 3
    # Scattering vectors
    if omega is None:
        s1 = peaks_xyz
    else:
        # scattering_vectors
        if len(omega) != len(peaks_xyz[0]):
            raise Exception(
                "omega and peaks arrays must have same number of peaks")
        s1 = peaks_xyz - compute_grain_origins(omega, wedge, chi,
                                               t_x, t_y, t_z)
    # CHANGED to HFP convention 4-9-2007
    eta = np.degrees(np.arctan2(-s1[1, :], s1[2, :]))
    s1_perp_x = np.sqrt(s1[1, :] * s1[1, :] + s1[2, :] * s1[2, :])
    tth = np.degrees(np.arctan2(s1_perp_x, s1[0, :]))
    return tth, eta


def compute_xyz_from_tth_eta(tth, eta, omega,
                             t_x=0.0, t_y=0.0, t_z=0.0,
                             #       == phi at chi=90
                             wedge=0.0,  # Wedge == theta on 4circ
                             chi=0.0,  # == chi - 90
                             **kwds):  # last line is for laziness -
    """
    Given the tth, eta and omega, compute the xyz on the detector

    crystal_translation is the position of the grain giving rise
    to a diffraction spot
                 in x,y,z ImageD11 co-ordinates
                 x,y with respect to axis of rotation and or beam centre ??
                 z with respect to beam height, z centre

    omega data needed if crystal translations used
    """
    # xyz = unit vectors along the scattered vectors
    xyz = np.zeros((3, tth.shape[0]), np.float)
    rtth = np.radians(tth)
    reta = np.radians(eta)
    xyz[0, :] =  np.cos(rtth)
    #  eta = np.degrees(np.arctan2(-s1[1, :], s1[2, :]))
    xyz[1, :] = -np.sin(rtth) * np.sin(reta)
    xyz[2, :] =  np.sin(rtth) * np.cos(reta)

    # Find vectors in the fast, slow directions in the detector plane
    pks = np.array([(1, 0),
                    (0, 1),
                    (0, 0) ], np.float).T
    dxyzl = compute_xyz_lab(pks, **kwds)
    # == [xpos, ypos, zpos] shape (3,n)
    #
    # This was based on the recipe from Thomas in Acta Cryst ...
    #  ... Modern Equations of ...

    ds = dxyzl[:,0] - dxyzl[:,2]  # 1,0 in plane is (1,0)-(0,0)
    df = dxyzl[:,1] - dxyzl[:,2]  # 0,1 in plane
    dO = dxyzl[:,2]               # origin pixel

    # Cross products to get the detector normal
    # Thomas uses an inverse matrix, but then divides out the determinant anyway
    det_norm = np.cross( ds, df )

    # Scattered rays on detector normal
    norm = np.dot( det_norm, xyz )
    # Check for divide by zero
    msk = (norm == 0)
    needmask = False
    if msk.sum()>0:
        norm += msk
        needmask = True

    # Intersect ray on detector plane
    sc = np.dot( np.cross( df, dO ), xyz ) / norm
    fc = np.dot( np.cross( dO, ds ), xyz ) / norm

    if (t_x != 0) or (t_y != 0) or (t_z != 0):
        go = compute_grain_origins(omega,
                                   wedge=wedge, chi=chi,
                                   t_x=t_x, t_y=t_y, t_z=t_z)
        # project these onto the detector face to give shifts
        sct = (  xyz * np.cross( df, go.T ).T ).sum(axis=0) / norm
        fct = (  xyz * np.cross( go.T, ds ).T ).sum(axis=0) / norm
        sc -= sct
        fc -= fct

    if needmask:
        fc = np.where( msk, 0, fc )
        sc = np.where( msk, 0, sc )

    return fc, sc


def compute_grain_origins(omega, wedge=0.0, chi=0.0,
                          t_x=0.0, t_y=0.0, t_z=0.0):
    """
    # print "Using translations t_x %f t_y %f t_z %f"%(t_x,t_y,t_z)
    # Compute positions of grains
    # expecting tx, ty, tz for each diffraction spot
    #
    # g =  R . W . k
    #  g - is g-vector w.r.t crystal
    #  k is scattering vector in lab
    #  so we want displacement in lab from displacement in sample
    #  shift =  W-1  R-1 crystal_translation
    #
    # R = ( cos(omega) , sin(omega), 0 )
    #     (-sin(omega) , cos(omega), 0 )
    #     (         0  ,         0 , 1 )
    #
    # W = ( cos(wedge) ,  0  ,  sin(wedge) )
    #     (         0  ,  1  ,          0  )
    #     (-sin(wedge) ,  0  ,  cos(wedge) )
    #
    # C = (         1  ,          0  ,       0     ) ??? Use eta0 instead
    #     (         0  ,   cos(chi)  ,  sin(chi)   )  ??? Use eta0 instead
    #     (         0  ,  -sin(chi)  ,  cos(chi)   )  ??? Use eta0 instead
    """
    w = np.radians(wedge)
    WI = np.array([[np.cos(w),         0, -np.sin(w)],
                   [0,           1,         0],
                   [np.sin(w),         0,  np.cos(w)]], np.float)
    c = np.radians(chi)
    CI = np.array([[1,            0,         0],
                   [0,     np.cos(c), -np.sin(c)],
                   [0,     np.sin(c),  np.cos(c)]], np.float)
    t = np.zeros((3, omega.shape[0]), np.float)  # crystal translations
    # Rotations in reverse order compared to making g-vector
    # also reverse directions. this is trans at all zero to
    # current setting. gv is scattering vector to all zero
    om_r = np.radians(omega)
    # This is the real rotation (right handed, g back to k)

    # Modified for translating sample, input t_x,t_y,t_z are
    # now precomputed scattering positions.
    #---------------------------------------------------
    t[0, :] = t_x
    t[1, :] = t_y
    t[2, :] = t_z
    
    # import matplotlib.pyplot as plt
    # plt.plot(t_x, t_y, "bo")
    # plt.plot(np.cos(om_r) * t_x - np.sin(om_r) * t_y, np.sin(om_r) * t_x + np.cos(om_r) * t_y, "ro")
    # plt.xlim([-600,600])
    # plt.ylim([-600,600])
    # plt.show()

    #old code:
    #t[0, :] = np.cos(om_r) * t_x - np.sin(om_r) * t_y
    #t[1, :] = np.sin(om_r) * t_x + np.cos(om_r) * t_y
    #t[2, :] = t_z
    #---------------------------------------------------


    if chi != 0.0:
        c = np.cos(np.radians(chi))
        s = np.sin(np.radians(chi))
        u = np.zeros(t.shape, np.float)
        u[0, :] = t[0, :]
        u[1, :] = c * t[1, :] + -s * t[2, :]
        u[2, :] = s * t[1, :] + c * t[2, :]
        t = u
    if wedge != 0.0:
        c = np.cos(np.radians(wedge))
        s = np.sin(np.radians(wedge))
        u = np.zeros(t.shape, np.float)
        u[0, :] = c * t[0, :] + -s * t[2, :]
        u[1, :] = t[1, :]
        u[2, :] = s * t[0, :] + c * t[2, :]
        t = u
    return t

def compute_k_vectors(tth, eta, wvln):
    """
    generate k vectors - scattering vectors in laboratory frame

    NOTE: These are the G_l vectors. not k vectors -... - Axel Henningsson 2021
          These do not have length 2pi/wvlv but some other scaling.
    """
    tth = np.radians(tth)
    eta = np.radians(eta)
    c = np.cos(tth / 2)  # cos theta
    s = np.sin(tth / 2)  # sin theta
    ds = 2 * s / wvln
    k = np.zeros((3, tth.shape[0]), np.float)
    # x - along incident beam
    k[0, :] = -ds * s  # this is negative x
    # y - towards door
    k[1, :] = -ds * c * np.sin(eta)  # CHANGED eta to HFP convention 4-9-2007
    # z - towards roof
    k[2, :] = ds * c * np.cos(eta)

    return k


def compute_g_vectors(tth,
                      eta,
                      omega,
                      wvln,
                      wedge=0.0,
                      chi=0.0):
    """
    Generates spot positions in reciprocal space from
      twotheta, wavelength, omega and eta
    Assumes single axis vertical
    ... unless a wedge angle is specified
    """
    k = compute_k_vectors(tth, eta, wvln)
#    print k[:,0]
    return compute_g_from_k(k, omega, wedge, chi)


def compute_g_from_k(k, omega, wedge=0, chi=0):
    """
    Compute g-vectors with cached k-vectors
    """
    om = np.radians(omega)
    # G-vectors - rotate k onto the crystal axes
    g = np.zeros((3, k.shape[1]), np.float)
    t = np.zeros((3, k.shape[1]), np.float)
    #
    # g =  R . W . k where:
    # R = ( cos(omega) , sin(omega), 0 )
    #     (-sin(omega) , cos(omega), 0 )
    #     (         0  ,         0 , 1 )
    #
    # W = ( cos(wedge) ,  0  ,  sin(wedge) )
    #     (         0  ,  1  ,          0  )
    #     (-sin(wedge) ,  0  ,  cos(wedge) )
    #
    # C = (         1  ,         0  ,      0     )
    #     (         0  ,  cos(chi)  , sin(chi)   )
    #     (         0  , -sin(chi)  , cos(chi)   )
    #
    if wedge != 0.0:
        c = np.cos(np.radians(wedge))
        s = np.sin(np.radians(wedge))
        t[0, :] = c * k[0, :] + s * k[2, :]
        t[1, :] = k[1, :]
        t[2, :] = -s * k[0, :] + c * k[2, :]
        k = t.copy()
    if chi != 0.0:
        c = np.cos(np.radians(chi))
        s = np.sin(np.radians(chi))
        t[0, :] = k[0, :]
        t[1, :] = c * k[1, :] + s * k[2, :]
        t[2, :] = -s * k[1, :] + c * k[2, :]
        k = t.copy()
    # This is the reverse rotation (left handed, k back to g)
    g[0, :] = np.cos(om) * k[0, :] + np.sin(om) * k[1, :]
    g[1, :] = -np.sin(om) * k[0, :] + np.cos(om) * k[1, :]
    g[2, :] = k[2, :]
    return g


def uncompute_g_vectors(g, wavelength, wedge=0.0, chi=0.0):
    """
    Given g-vectors compute tth,eta,omega
    assert uncompute_g_vectors(compute_g_vector(tth,eta,omega))==tth,eta,omega
    """
    if wedge == chi == 0:
        post = None
    else:
        post = gv_general.wedgechi( wedge=wedge, chi=chi )
    omega1, omega2, valid = gv_general.g_to_k(
        g, wavelength,axis=[0,0,-1], pre=None, post=post )
    # we know g, omega. Compute k as ... ?
    if post is None:
        pre = None
    else:
        pre = gv_general.chiwedge( wedge=wedge, chi=chi ).T
    k_one = gv_general.k_to_g( g, omega1, axis=[0,0,1],
                               pre = pre, post=None)
    k_two = gv_general.k_to_g( g, omega2, axis=[0,0,1],
                               pre = pre, post=None)
    #
    # k[1,:] = -ds*c*sin(eta)
    # ------    -------------   .... tan(eta) = -k1/k2
    # k[2,:] =  ds*c*cos(eta)
    #
    eta_one = np.arctan2(-k_one[1, :], k_one[2, :])
    eta_two = np.arctan2(-k_two[1, :], k_two[2, :])
    #
    #
    ds = np.sqrt(np.sum(g * g, 0))
    s = ds * wavelength / 2.0  # sin theta
    tth = np.degrees(np.arcsin(s) * 2.) * valid
    eta1 = np.degrees(eta_one) * valid
    eta2 = np.degrees(eta_two) * valid
    omega1 = omega1 * valid
    omega2 = omega2 * valid
    return tth, [eta1, eta2], [omega1, omega2]


from __future__ import print_function
import numpy as np
from skimage.transform import iradon, radon
import sys
import matplotlib.pyplot as plt


def FBP_slice( grains, flt, omegastep, rcut, ymin, ystep, number_y_scans):
    grain_masks=[]
    grain_recons=[]
    for i,g in enumerate(grains):
        sinoangles, sino, recon = FBP_grain( g, flt, \
                    ymin, ystep, omegastep, number_y_scans )
        normalised_recon = recon/recon.max()
        grain_recons.append(normalised_recon)
        mask = normalised_recon > rcut
        grain_masks.append(mask)
    update_grainshapes(grain_recons,grain_masks)
    return grain_masks


def FBP_grain( g, flt, ymin, ystep, omegastep, number_y_scans ):
    """
    Reconstruct a 2d grain shape from diffraction data using Filtered-Back-projection.
    """

    # Measured relevant data for the considered grain
    omega = flt.omega[ g.mask ].copy()
    dty = flt.dty[g.mask].copy()
    sum_intensity = flt.sum_intensity[g.mask].copy()

    if np.min(omega)<0 and np.max(omega)>0:
        # Case 1: -180 to 180 rotation and positive y-scans we make
        # a transformation back into omega=[0 180] in three steps:

        # (I) Half the intensity for peaks entering the sinogram twice.
        doublets_mask = dty<=np.abs(np.min(dty))
        sum_intensity[doublets_mask] = sum_intensity[doublets_mask]*(1/2)

        # (II) Map the negative omega values back to positive values, [0 180]
        omega_mask = omega < 0
        omega[omega_mask] = omega[omega_mask] + 180

        # (III) Flip the sign of the y-scan coordinates with negative omega values.
        dty[omega_mask]   = -dty[omega_mask]

    elif np.min(omega)>=0 and np.max(omega)<=180:
        # Case 2: 0 to 180 rotation and both negative and positive y-scans
        # nothing needs be done since this is the standrad tomography case
        pass
    else:
        raise ValueError("Scan configuration not implemented. omega=0,180 scans with positive and \
                          negative y-translations and omega=-180,180 scans with all positive      \
                          y-translations are supported.")

    # Angular range into which to bin the data (step in sinogram)
    angular_bin_size = 180./(number_y_scans)

    # Indices in sinogram for the y-scan and angles
    iy = np.round( (dty - ymin) / ystep ).astype(int)
    iom = np.round( omega / angular_bin_size ).astype(int)

    # Build the sinogram by accumulating intensity
    sinogram = np.zeros( ( number_y_scans, np.max(iom)+1 ), np.float )
    for i,I in enumerate(sum_intensity):
        dty_index   = iy[i]
        omega_index = iom[i]
        sinogram[dty_index, omega_index] += I

    # Normalise the sinogram to account for the intensity not being proportional
    # only to density but also to eta and theta and a lot of other stuff.
    normfactor = sinogram.max(axis=0)

    normfactor[normfactor==0]=1.0
    sinogram = sinogram/normfactor

    # Perform reconstruction by inverse radon transform of the sinogram
    theta = np.linspace( angular_bin_size/2., 180. - angular_bin_size/2., sinogram.shape[1] )
    back_projection = iradon( sinogram, theta=theta, output_size=number_y_scans, circle=True )


    return [], sinogram, back_projection 

def update_grainshapes( grain_recons, grain_masks):
    '''
    Update a set of grain masks based on their overlap and intensity.
    At each point the grain with strongest intensity is assigned

    Assumes that the grain recons have been normalized
    '''

    for i in range(grain_recons[0].shape[0]):
        for j in range(grain_recons[0].shape[1]):
            if conflict_exists(i,j,grain_masks):
                max_int = 0.0
                leader  = None
                for n,grain_recon in enumerate(grain_recons):
                    if grain_recon[i,j]>max_int:
                        leader = n
                        max_int = grain_recon[i,j]

                #The losers are...
                for grain_mask in grain_masks:
                    grain_mask[i,j]=0

                #And the winner is:
                grain_masks[leader][i,j]=1

def conflict_exists( i,j,grain_masks):
    '''
    Help function for update_grainshapes()
    Checks if two grain shape masks overlap at index i,j
    '''
    claimers = 0
    for grain_mask in grain_masks:
        if grain_mask[i,j]==1:
            claimers+=1
    if claimers>=2:
        return True
    else:
        return False


import numpy as np
import matplotlib.pyplot as plt
import shapely.geometry
from scipy.ndimage.morphology import binary_dilation

def voxels_to_polygon( image_stack, pixel_size, center=(0.5,0.5) ):
    '''Take a stack of images and produce a stack of shapely polygons.
    The images are interpreted as a solid shape with boundary along the pixel 
    exterior edge. Thus an image eith a single nonzero pixel will return a square 
    polygon with sidelength equal to the pixel_size.

    IN:
        image_stack: list of binary (1.0,0) numpy array 2d images each depicting
                     a single connected region of 1.0 surrounded by 0.0.
        pixel_size: The absolute pixel size of the input images. Used to make the
                    output polygons coordinates real spaced.
        center: the relative origin of the image, axis=0 is x and axis=1 is y
                increasing with increasingf index. For instance center=(0.5,0.5)
                will select the centre of the image as the orign.

    OUT:
        polygon_stack: list of shapely.geometry.polygons each representing the bound
                       of the corresponding input binary image.
    '''
    polygon_stack = [pixels_to_polygon(image, pixel_size, center) for image in image_stack]
    return polygon_stack

def pixels_to_polygon( image, pixel_size, center=(0.5,0.5) ):
    '''Take a single image and produce a shapely polygon.
    '''
    expanded_image = expand_image(image, factor=3)
    indices = get_image_boundary_index( expanded_image )
    coordinates = indices_to_coordinates(indices, pixel_size/3., center, expanded_image)
    polygon = shapely.geometry.Polygon(coordinates)
    #show_polygon_and_image(polygon, image, pixel_size, center) #<= DEBUG
    return polygon

def expand_image(image, factor):
    '''Expand 2d binary image so that each pixel is split by copying 
    into factor x factor number of pixels.
    '''
    expanded_image = np.repeat(image, factor, axis=1)
    expanded_image = np.repeat(expanded_image, factor, axis=0)
    return expanded_image

def get_image_boundary_index( image ):
    '''Find the pixel indices of the boundary pixels of a binary image.
    '''

    boundary_image = get_boundary_image( image )

    bound_indx = np.where( boundary_image==1 )
    ix,iy = bound_indx[0][0],bound_indx[1][0] # starting index
    indices = [(ix,iy)]
    while( not len(indices)==np.sum(boundary_image) ):
        # Walk around border and save boundary pixel indices
        mask = np.zeros(boundary_image.shape)
        mask[ np.max([0,ix-1]):ix+2, iy ] = 1
        mask[ ix, np.max([iy-1]):iy+2 ]   = 1
        neighbour_indx = np.where( boundary_image*mask )
        for ix,iy in zip( neighbour_indx[0], neighbour_indx[1]):
            if (ix,iy) not in indices:
                indices.append( (ix,iy) )
                break
    indices = sparse_indices( indices )
    return indices

def get_boundary_image( image ):
    '''Return a pixel image with 1 along the boundary if the assumed
    object in image.
    '''
    k = np.ones((3,3),dtype=int)
    dilation = binary_dilation( image==0, k, border_value=1 )
    boundary_image = dilation*image
    return boundary_image

def sparse_indices( indices ):
    '''Remove uneccesary nodes in the polygon (three nodes on a line is uneccesary).
    '''
    new_indices = []
    for i in range(0, len(indices)-1):
        if not (indices[i-1][0]==indices[i][0]==indices[i+1][0] or \
        indices[i-1][1]==indices[i][1]==indices[i+1][1]):
            new_indices.append(indices[i])
    return new_indices

def indices_to_coordinates(indices, pixel_size, center, image  ):
    '''Compute real space coordinates of image boundary form set of pixel indices. 
    '''
    dx = image.shape[1]*center[0]
    dy = image.shape[0]*center[1]
    coordinates = []
    for c in indices:
        # Verified by simulated nonsymmetric grain
        ycoord = pixel_size*(  c[1] + 0.5 - dx + (c[1]%3 - 1)*0.5 )
        xcoord = pixel_size*( -c[0] - 0.5 + dy - (c[0]%3 - 1)*0.5 )
        coordinates.append( (xcoord,ycoord) )
    return coordinates


def get_integral_paths( angles, ytrans, zpos, sample_polygon, show_geom=False ):
    '''Compute entry-exit points for a scanrange.
    '''

    # Instantiate lists to contain all measurements
    all_entry, all_exit, all_nhat, all_L, all_nsegs, bad_lines = [],[],[],[],[],[]
    xray_endpoints = get_xray_endpoints( sample_polygon )

    # Loop over all experimental settings
    for i,(ang,dty) in enumerate( zip( angles, ytrans ) ):

        # Translate and rotate the xray endpoints according to ytrans and angle
        c, s = np.cos( np.radians( -ang ) ), np.sin( np.radians( -ang ) )
        rotz = np.array([[c,-s],[s,c]])
        rx = rotz.dot( xray_endpoints + np.array([[0,0],[dty,dty]]) ) 
        xray_polygon = shapely.geometry.LineString( [ rx[:,0], rx[:,1] ] )

        # compute the intersections between beam and sample in sample coordinates
        intersection_points = get_intersection( xray_polygon, sample_polygon, zpos  )
    
        if intersection_points is None:
            # If a measurement missed the sample or graced a corner, we skipp ahead
            bad_lines.append(i)
            continue
        else:
            # make a measurement at the current setting
            entry, exit, nhat, L, nsegs = get_quanteties( intersection_points )

            # save the measurement results in global lists
            all_entry.append( entry )
            all_exit.append( exit )
            all_nhat.append( nhat )
            all_L.append( L )
            all_nsegs.append( nsegs )

            if show_geom:
                # Show sample and xray polygons
                print('entry ',entry)
                print('exit ',exit)
                print('nhat ',nhat)
                print('L ',L)
                xc,yc = xray_polygon.xy
                plt.figure(figsize=(11,8))
                plt.scatter(entry[0::3], entry[1::3],c='k',zorder=200,label='entry')
                plt.scatter(exit[0::3], exit[1::3],c='b',zorder=200,label='exit')
                plt.plot(xc,yc,c='y',label='Beam')
                plt.arrow(0, 0, 20*nhat[0],20*nhat[1], head_width=2, color='r', zorder=100,label=r'$\hat{n}$')
                xc,yc = sample_polygon.exterior.xy
                plt.fill(xc,yc,color='gray',label='Grain',zorder=1)
                plt.title('L='+str(L)+' angle='+str(ang)+r'$^o$,   dty='+str(dty)+',   z_translation='+str(zpos)+ 'nsegs='+str(nsegs))
                plt.axis('equal')
                plt.xlabel('x')
                plt.ylabel('y')
                
                xcircle = np.linspace(-dty, dty, 100)
                ycircle = np.sqrt( dty**2 - xcircle**2 )
                plt.plot( xcircle, ycircle, c='g' )
                plt.plot( xcircle, -ycircle, c='g', label='circle with R=dty' )
                plt.grid(True)
                plt.legend()
                plt.show()

    # repack lists of measurements into numpy arrays of desired format
    entry, exit, nhat, L, nsegs = repack(all_entry, all_exit, all_nhat, all_L, all_nsegs)

    return  entry, exit, nhat, L, nsegs, bad_lines


def get_xray_endpoints( sample_polygon ):
    '''Calculate endpoitns of xray line segement. The lenght of the 
    line segment is adapted to make sure xray always convers the full
    length of the sample.
    '''
    xc, yc = sample_polygon.exterior.xy
    xmin = np.min( xc )
    xmax = np.max( xc )
    ymin = np.min( yc )
    ymax = np.max( yc )
    D = np.sqrt( (xmax-xmin)**2 + (ymax-ymin)**2 )
    return np.array([ [-1.1*D, 1.1*D], [0,0] ])

def get_intersection( xray_polygon, sample_polygon, z ):
    '''Compute the 3d coordinates of intersection between xray and
    sample.
    '''
    intersection = sample_polygon.intersection( xray_polygon )
    if intersection.is_empty or isinstance(intersection, shapely.geometry.point.Point):
        # we missed the sample with the beam
        intersection_points = None
    elif isinstance(intersection, shapely.geometry.linestring.LineString):
        # we got a single line segment intersection
        intersection_points = np.zeros( (2,3) )
        intersection_points[:2,:2] = np.array( intersection.xy ).T
        intersection_points[:,2] = z
    elif isinstance(intersection, shapely.geometry.multilinestring.MultiLineString):
        # we got multiple line segments intersection
        intersection_points = np.zeros( (2*len(intersection.geoms),3) )
        for i,line_segment in enumerate(intersection.geoms):
            intersection_points[2*i:2*(i+1),:2] = np.array( line_segment.xy ).T
        intersection_points[:,2] = z 
    return intersection_points

def get_quanteties( intersection_points ):
    nsegs = intersection_points.shape[0]//2
    entry,exit = [],[]
    p1 = intersection_points[0,:] 
    p2 = intersection_points[1,:]
    nhat = list( (p2-p1)/np.linalg.norm(p2-p1) )

    L = 0
    for i in range( nsegs ):
        p1 = intersection_points[2*i,:] 
        p2 = intersection_points[2*i+1,:]
        entry.extend( list(p1) )
        exit.extend( list(p2) )
        length = np.linalg.norm(p2-p1)
        L += length

    return entry, exit, nhat, L, nsegs

def repack( all_entry, all_exit, all_nhat, all_L, all_nsegs ):
    '''Repack global measurement list into numpy arrays of desired format.
    '''
    N = len( all_L )

    p = max( max(all_nsegs), 1 )
    nsegs = np.array(all_nsegs).reshape( 1,N )
    L = np.array(all_L).reshape( 1,N )

    entry = np.zeros( (3*p, N) )
    for i,en in enumerate(all_entry):
        entry[:len(en[:]),i] = en[:]

    exit = np.zeros( (3*p, N) )
    for i,ex in enumerate(all_exit):
        exit[:len(ex[:]),i] = ex[:]

    nhat = np.array( all_nhat ).T
    return entry, exit, nhat, L, nsegs

def show_polygon_and_image( polygon, image, pixel_size, center ):
    '''Plot a image and polygon for debugging purposes
    '''
    fig,ax = plt.subplots(1, 2, figsize=(12,6))
    fig.suptitle('Center at '+str(center))
    xc,yc = polygon.exterior.xy
    xcenter = image.shape[1]*pixel_size*center[0]
    ycenter = image.shape[0]*pixel_size*center[1]
    ax[0].imshow(image,cmap='gray')
    ax[0].set_title('Pixel image')
    ax[0].arrow( int(image.shape[1]*center[0]), int(image.shape[0]*center[1]), \
        image.shape[0]//4, 0, color='r', head_width=0.15 ) # y
    ax[0].text( int(image.shape[1]*center[0])+image.shape[1]//4, int(image.shape[0]*center[1])+0.25, \
        'y',color='r' )
    ax[0].arrow( int(image.shape[1]*center[0]), int(image.shape[0]*center[1]), \
        0, -image.shape[1]//4, color='r', head_width=0.15 ) # x
    ax[0].text( int(image.shape[1]*center[0])+0.25, int(image.shape[0]*center[1])-image.shape[1]//4,  \
        'x',color='r' )
    ax[1].set_title('Polygon representation')
    ax[1].fill(xc,yc,c='gray',zorder=1)
    ax[1].scatter(xc,yc,c='r',zorder=2)
    ax[1].grid(True)
    ax[1].scatter(0, 0 , c='b',zorder=3)
    ax[1].set_xlim([ -xcenter, image.shape[1]*pixel_size - xcenter ])
    ax[1].set_ylim([ -ycenter, image.shape[0]*pixel_size - ycenter ])
    ax[1].set_xlabel('x')
    ax[1].set_ylabel('y')
    plt.show()




import numpy as np
import matplotlib.pyplot as plt
from ImageD11 import columnfile, parameters, grain
from s3dxrd.utils import peak_mapper, reconstruct_grainshapes

class RawMeasurements(object):
    """Container of semi-raw diffraction data and grain properties.

    Performs diffraction peak mapping and grain topology reconstruction
    working of ImageD11 file formats. These files are already proccessed in
    the sense that the grains have been indexed and detector tilt has
    been calibrated for etc. In scanning-3DXRD each layer (x-y) is scanned
    independently of z translations. Thus the input is a stacks of
    measurements, one file per sample z-translation. The coordinate system
    and experimental setup is described in:

        Henningsson, N. A., Hall, S. A., Wright, J. P. & Hektor, J. 
        (2020). J. Appl. Cryst. 53, 314-325.
        https://doi.org/10.1107/S1600576720001016

    Attributes:
        zpos (list of float): Sample z-translations corresponding to grain_slices
            indices and peak_stack indices.
        grain_slices (list of lists of :obj:`grain`): Loaded ImageD11 grain '.map'
            or '.ubi' files. Each z-translation has a list of grains present in
            the slice. The map files are loaded via the ImageD11 grain module.
        peak_stack (list of :obj:`columnfile`): Loaded ImageD11 peak files. Each
            instance of the peak stack list contains a columnfile describing
            diffraction peaks originating from correpsonding z-translation. The 
            peak files are loaded via the ImageD11 columnfile module.
        params (:obj:`parameters`): Loaded ImageD11 parameter file. This object
            gives access to meta parameters of the experiment, such as detector tilt,
            sample to detector distance, and crystal cell parameters. The parameter
            file is loaded via the ImageD11 parameters module.
        omegastep (float): Rotation stage readout intervall in degrees. Each recorded
            frame in the experiment is was integrated over this intervall.
        ystep (float): y-translation stepsize between measurements. Normally the same
            as the beamsize.
        number_y_scans (float): The number of scanned y-positions made in each z-layer.
        ymin (float): Minium y-coordinate seen by the detector. These values differ from
            what is in the orginal columnfile in the case when the columnfile do not have
            a symmetric scanning intervall over y=0.
        ymax (float): Maximum y-coordinate seen by the detector. These values differ from
            what is in the orginal columnfile in the case when the columnfile do not have
            a symmetric scanning intervall over y=0.
        grain_topology_mask (list of 2D numpy arrays): Grain topology in binarized form.
            Each instacne of the list contains the grain shape at the corresponding
            z-slice. This attribute starts as None, and is set as an result of an active
            action.
    """

    def __init__(self, 
                flt_paths,
                zpos,
                param_path,
                ubi_paths,
                omegastep, 
                ymin, 
                ymax, 
                ystep):
        """Container of semi-raw diffraction data and grain properties.

        Args:
            zpos (list of float): Sample z-translations corresponding to flt_paths.
            ubi_paths (str): Absolute paths to ID11 grain files.
            flt_paths (list of str): Absolute paths to ID11 column files.
            param_path (str): Absolute path to ID11 parameter file.
            omegastep (float): Rotation stage readout intervall in degrees.
            ystep (float): stepsize of sample y-translation between measurements
            ymin (float): Minium y-coordinate seen by the detector, 
                as given in ID11 column file.
            ymax (float): Maximum y-coordinate seen by the detector, 
                as given in ID11 column file.
        """
        self.zpos = zpos
        self.grain_slices = [ grain.read_grain_file( ubip ) for ubip in ubi_paths ]
        self.peak_stack = [columnfile.columnfile(flt) for flt in flt_paths]
        self.tot_nbr_peaks = sum( [peaks.nrows for peaks in self.peak_stack] )
        self.params = parameters.read_par_file( param_path )
        self.omegastep = omegastep
        self.ystep = ystep
        self.number_y_scans = np.round((abs(ymax-ymin)/ystep)).astype(int)+1
        for p in self.peak_stack: 
            p.dty[:] = p.dty - ymin - self.ystep*(self.number_y_scans//2)
        self.ymin = -ystep*(self.number_y_scans//2)
        self.ymax =  ystep*(self.number_y_scans//2)
        self.grain_topology_mask = None

        self.params.parameters['ystep'] = self.ystep
        self.params.parameters['ymin'] = self.ymin
        self.params.parameters['t_x'] = None
        self.params.parameters['t_y'] = None
        self.params.parameters['t_z'] = None

    def map_peaks(self, hkltol, nmedian ):
        """Perform Grain centroid refinement and diffraction peak to grain mappings.

        This method refines the semi-raw diffraction data contained by the input 
        peak files. The grains of self.grain_slices will be given masks and refined
        average orientation and strain. The columnfiles of the peak_stack will be
        recomputed based on the centroid of the claimer grain.

        Note:
            This function mutates the attributes of the class and has no return.

        Args:
            hkltol (float): Tolerance on Miller indices (hkl) to match a peak.
            nmedian (float): Outlier threshold. If any of the diffraction angles 
                (2*theta, omega, eta) are more than nmedian times deviating from the 
                median deviation (between model and measurement) the peak is considered
                an outlier and removed.
        """
        for gs, flt, dtz in zip( self.grain_slices, self.peak_stack, self.zpos ):
            peak_mapper.map_peaks(flt, gs, self.params, self.omegastep, \
                                        hkltol, nmedian, self.ymin, self.ystep, self.number_y_scans)

    def reconstruct_grain_topology(self, rcut):
        """Perform Filtered Back Projection to approximate grain shapes.

        The recorded intensity of the diffraction peaks are used to create
        a grain topology map. Each grain is reconstructed on a pixelated grid,
        and thresholded to a binary 1 - 0, grain or no grain, map.

        Note:
            This function mutates the attributes of the class and has no return.

        Args:
            rcut (float): Threshold for segementing out grain. Reconstructed 
            relative intensity of less than rcut is considered void, and intensity
            higher than rcut is considered grain.
        """
        self.grain_topology_mask=[]
        for gs, flt, dtz in zip( self.grain_slices, self.peak_stack, self.zpos ):
            grain_topology_mask = reconstruct_grainshapes.FBP_slice(gs, flt,       \
                                        self.omegastep, rcut, self.ymin, \
                                        self.ystep, self.number_y_scans)                   
            self.grain_topology_mask.append( grain_topology_mask )



from pyevtk.hl import gridToVTK
import numpy as np
import pickle

def as_vtk_voxel_volume(file, arrays, names):
    """Save arrays as fields that are paraview compatible (vtk files).

    Args:
        file (:obj:`string`): Asolute path to save the output at ending with desired filename.
        arrays (:obj:`list` of :obj:`numpy array`): The arrays to be saved.
        names (:obj:`list` of :obj:`string`): Names of the arrays refered to by paraview. These
            will also be the default titles in paraview.

    """

    x = np.arange(0, arrays[0].shape[0]+1, dtype=np.int32)
    y = np.arange(0, arrays[0].shape[1]+1, dtype=np.int32)
    z = np.arange(0, arrays[0].shape[2]+1, dtype=np.int32)

    cellData = {}
    for array,name in zip(arrays,names):
        cellData[name] = array

    gridToVTK(file, x, y, z, cellData)

def save_object( path, object_to_save ):
    with open(path, "wb") as f:
        pickle.dump(object_to_save, f, pickle.HIGHEST_PROTOCOL)

def load_vectors(path):
    """Load vectors produced and saved by s3dxrd.Id11.peaks_to_vectors().

    Args:
        path (:obj:`string`): Asolute path ending with filename of file to load.

    Returns:
        (:obj:`dict`): dictionary with fields as specified in s3dxrd.Id11.peaks_to_vectors().

    """
    with open(path, 'rb') as f:
        return pickle.load(f)



import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), "utils/"))

import numpy as np
import matplotlib.pyplot as plt
from . import raw_measurements as raw_measurements
from s3dxrd.utils import topology as top 
from s3dxrd.utils import measurement_converter as mc
from s3dxrd.utils import save
import pickle
from s3dxrd.utils import slice_matcher
from xfab import tools
from scipy.ndimage.morphology import binary_opening

#TODO: Split this function into parts:
    # I)   Peak grain mapping + shape reconstruction.
    # II)  Cross slice mapping
    # III) Polygon representations
    # IV)  Parametric integrals
    # V)   Average strains and directions.

def peaks_to_vectors(flt_paths,
                    zpos,
                    param_path,
                    ubi_paths,
                    omegastep, 
                    ymin, 
                    ymax, 
                    ystep,
                    hkltol = 0.05,
                    nmedian = 5,
                    rcut = 0.2,
                    save_path=None ):
    """Convert an x-ray diffraction dataset saved in the ImageD11 format to vector format
    
    Based on an inital guess of ubi matrices all data contained in a sereis of Id11 peak files
    is analysed. Grian shapes are reconstructed and grain orientations refined. The data is converted
    into a vector format where each measurement is associated to a parametric line trhough the grain
    as well as an average approximate strain along the diffracting planes.

    The proccessed output from this function can be used to further perform regression for strain tensor
    maps using any of the in package supported regression techinques.

    Args:
        flt_paths (:obj:`list` of :obj:`string`): Absolute file paths to Id11 peaks files. These must contain
            a column named ```dty``` that stores the sample y-translations in units of microns. These translations
            should be centered at the rotation axis (dty=0 means the xray go through the rotation axis.)
        zpos (:obj:`list` of :obj:`float`): z-translations of sample corresponding to peaks in :obj:`flt_paths`.
        param_path (:obj:`string`): Absolute file path to Id11 parameters file.
        ubi_paths (:obj:`list` of :obj:`string`):  Absolute file paths to Id11 ubi matrices file.
        omegastep (:obj:`float`): Rotation intervall in degrees over wich detector images are where read out.
        ymin (:obj:`float`): Minimum y-translation of smaple in units of microns.
        ymax (:obj:`float`): Maximum y-translation of smaple in units of microns. 
        ystep (:obj:`float`): Sample y-translation stepsize.
        hkltol (:obj:`float`): Miller index tolerance for assigning peaks to a grain. Defaults to 0.05.
        nmedian (:obj:`float`): Number of median deviations of a peak to discard as outlier. Defaults to 5.
        rcut (:obj:`float`):  Relative threshold for defining grain shape from tomographic reconstrction. Defaults to 0.2.
        save_path (:obj:`string`): Path at which to save all computed for arrays. Defaults to :obj:`None`.

    Returns:
        (:obj:`dict`): dictionary with keys and values:

            **Y** (:obj:`numpy array`): Average scalar strains along scanned lines. ``shape=(N,)``
  
            **sig_m** (:obj:`numpy array`): Per measurement standard deviations. ``shape=(N,)``
  
            **entry** (:obj:`numpy array`): Entry points for line integral meaurements.  ``shape=(k,N)``
  
            **exit** (:obj:`numpy array`): Exit points for line integral meaurements.  ``shape=(k,N)``
  
            **nhat** (:obj:`numpy array`): X-ray beam direction normal in sample coordinate system. ``shape=(3,N)``
  
            **kappa** (:obj:`numpy array`): Direction of strain (Y) in sample coordinate system. ``shape=(3,N)``
  
            **L** (:obj:`numpy array`): Lengths of scanned lines. ``shape=(N,)``
  
            **nsegs** (:obj:`numpy array`): Number of segments for each line integral. ``shape=(N,)``
 
            **polygons** (:obj:`dict` of :obj:`dict` of :obj:`Shapely Polygons`): Polygonal boundary representation of grians.
  
            **polygon_orientations** (:obj:`dict` of :obj:`dict` of :obj:`numpy array`): Crystal orientation of each grain slice in polygons.
  
            **orientations** (:obj:`numpy array`): Per measurement crystal orientation (uniform along scan-line).
  
            **measurement_grain_map** (:obj:`numpy array`): dictionary keys as integers corresponding to polygons, such that
            ``polygons[str(measurement_grain_map[i])]`` gives the polygon slices dictionary of the grain
            that gave rise to measurement number i. ``shape=(N,)``

            **labeled_volume** (:obj:`numpy array`): 3d numpy array of sample with values corresponding to grain index.

            **labeled_grains** (:obj:`dict` of :obj:`dict` of :obj:`ImageD11 Grain`): Id11 grain for each grain slice in polygons.

    """

    print('Initiating data conversion from Id11 format to gpxrd format...')
    rm = raw_measurements.RawMeasurements(flt_paths, zpos, param_path, ubi_paths, omegastep, ymin, ymax, ystep)

    print('Mapping a total of '+str(rm.tot_nbr_peaks)+' peaks...')
    rm.map_peaks( hkltol, nmedian )

    print('Reconstructing grain topologies...')
    rm.reconstruct_grain_topology( rcut )

    # Cleanup the recons from floating pixels.
    
    for i in range(len(rm.grain_topology_mask)):
        for j in range(len(rm.grain_topology_mask[i])):
            rm.grain_topology_mask[i][j] = binary_opening(rm.grain_topology_mask[i][j], structure=np.ones((3,3)))

    # Cross slice mapping of grains, giving each grain a unique label so it can be tracked across z-slices.
    print('Cross slice mapping grains...')
    labeled_volume, labeled_grains = slice_matcher.match_and_label( rm.grain_topology_mask, rm.grain_slices )

    print('Converting diffraction peaks to GP-XRD quanteties...')
    all_Y, all_sig_m, all_entry, all_exit, all_nhat, all_kappa, all_L, all_nsegs = [],[],[],[],[],[],[],[]
    orientations = None
    measurement_grain_map = []

    polygons = {}
    polygon_orientations =  {}
    for grain_indx in np.unique( labeled_volume[labeled_volume>0] ):
        polygons[str(grain_indx)] = {}
        polygon_orientations[str(grain_indx)] = {}

    for i,(zpos, peaks) in enumerate(zip( rm.zpos, rm.peak_stack)):

        print( "Working on z-position: " + str(zpos) )

        for grain_indx in np.unique( labeled_volume[:,:,i] ):

            if grain_indx==0:
                continue

            # Each grain slice is represented as a 2d polygon
            image = labeled_volume[:,:,i]==grain_indx
            grain = labeled_grains[str(grain_indx)][str(i)]
            sample_polygon = top.voxels_to_polygon( [image], pixel_size=ystep, center=(0.5,0.5) )[0]

            if 0: # DEBUG:
                top.show_polygon_and_image( sample_polygon, image, pixel_size=ystep, center=(0.5,0.5) )

            measurements = mc.convert_measurements( rm.params, grain, peaks, rm.ymin, rm.ystep, rm.omegastep )
            
            Yz     =  measurements['strain']
            kappaz =  measurements['kappa']
            sig_mz =  measurements['sig_m']
            angles =  measurements['omega']
            dty    =  measurements['dty']

            entryz, exitz, nhatz, Lz, nsegsz, bad_lines = top.get_integral_paths( angles, dty, zpos, \
                                                                    sample_polygon, show_geom=False )

            # remove line integrals that missed sample
            Yz     = np.delete(Yz, bad_lines)
            kappaz = np.delete(kappaz, bad_lines, axis=0)
            sig_mz = np.delete(sig_mz, bad_lines)

            # Build per line integral measurement orientation matrix (Nx3x3)
            U = np.zeros((1,3,3))
            U[0,:,:] = tools.ubi_to_u(grain.ubi)[:,:]
            slice_orientations = np.repeat(U, len(Yz), axis=0)
            if orientations is None:
                orientations = slice_orientations
            else:
                orientations = np.concatenate( [orientations, slice_orientations], axis=0 )

            measurement_grain_map.extend( [grain_indx]*len(Yz) )

            all_Y.extend( Yz )
            all_sig_m.extend( sig_mz )
            all_entry.append( entryz )
            all_exit.append( exitz )
            all_nhat.append( nhatz )
            all_kappa.append( kappaz )
            all_L.extend( Lz )
            all_nsegs.extend( nsegsz )

            polygons[str(grain_indx)][str(i)] = sample_polygon
            polygon_orientations[str(grain_indx)][str(i)] = U[0,:,:]

    Y, sig_m, entry, exit, nhat, kappa, L, nsegs = _repack(all_Y, all_sig_m, all_entry, all_exit, all_nhat, all_kappa, all_L, all_nsegs)
    measurement_grain_map = np.array(measurement_grain_map)

    vectors = { 'Y':Y, 
                'sig_m':sig_m, 
                'entry':entry, 
                'exit':exit, 
                'nhat':nhat, 
                'kappa':kappa, 
                'L':L, 
                'nsegs':nsegs, 
                'orientations':orientations, 
                'measurement_grain_map':measurement_grain_map, 
                'labeled_volume':labeled_volume,
                'polygons': polygons,
                'polygon_orientations': polygon_orientations }

    if save_path is not None:
        print('Writing results to disc at '+save_path+'  ...')
        save.save_object(save_path, vectors)

    return vectors


def _repack(all_Y, all_sig_m, all_entry, all_exit, all_nhat, all_kappa, all_L, all_nsegs):
    '''_repack global measurement list into numpy arrays of desired format.
    '''
    Y = np.array( all_Y )
    N = len(Y)
    
    p = np.max( [np.max(nsegs) for nsegs in all_nsegs] )

    nsegs = np.concatenate(all_nsegs).reshape( 1,N )
    sig_m  = np.array(all_sig_m).reshape( N, 1)
    L = np.concatenate(all_L).reshape( 1,N )

    entry = np.zeros( (3*p, all_entry[0].shape[1]) )
    entry[:all_entry[0].shape[0],:] = all_entry[0][:,:]
    for i in range(1,len(all_entry)):
        tmp = np.zeros( (3*p, all_entry[i].shape[1]) )
        tmp[:all_entry[i].shape[0],:] = all_entry[i][:,:]
        entry = np.concatenate( [entry,tmp], axis=1 )

    exit = np.zeros( (3*p, all_exit[0].shape[1]) )
    exit[:all_exit[0].shape[0],:] = all_exit[0][:,:]
    for i in range(1,len(all_exit)):
        tmp = np.zeros( (3*p, all_exit[i].shape[1]) )
        tmp[:all_exit[i].shape[0],:] = all_exit[i][:,:]
        exit = np.concatenate( [exit,tmp], axis=1 )

    nhat = np.concatenate( all_nhat, axis=1 )
    kappa = np.concatenate( all_kappa, axis=0 ).T
    return Y, sig_m, entry, exit, nhat, kappa, L, nsegs



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from skimage.transform import iradon, radon
from xfab import tools
import torch

def vectors_to_torch(vectors):
    """Convert numpy arrays in a vectors dictionary to torch double format.

    Args:
        vectors (:obj:`dict`): vectors dictionary as specified by s3dxrd.measurements.Id11.peaks_to_vectors()
    
    Returns:
        (:obj:`dict`): dictionary with same fields as ```vectors``` but with torch tensors replacing
            numby arrays.
    """
    vectors["Y"]      =  torch.from_numpy(vectors["Y"]).double()
    vectors["entry"]  =  torch.from_numpy(vectors["entry"]).double()
    vectors["exit"]   =  torch.from_numpy(vectors["exit"]).double()
    vectors["nhat"]   =  torch.from_numpy(vectors["nhat"]).double()
    vectors["kappa"]  =  torch.from_numpy(vectors["kappa"]).double()
    vectors["L"]      =  torch.from_numpy(vectors["L"]).double()
    vectors["nsegs"]  =  torch.from_numpy(vectors["nsegs"]).double()
    vectors["sig_m"]  =  torch.from_numpy(vectors["sig_m"]).double()
    return vectors

def sind( a ): return np.sin( np.radians( a ) )

def cosd( a ): return np.cos( np.radians( a ) )

def omega_matrix( om ):
    return np.array([[cosd(om),-sind(om),0],[sind(om),cosd(om),0],[0,0,1]])

def get_k(tth, eta, wavelength):
    return (2*np.pi/wavelength)*np.array([1, 0, 0])

def get_k_prime(tth, eta, wavelength):
    k1 = cosd(tth)
    k2 = -sind(tth)*sind(eta)
    k3 = sind(tth)*cosd(eta)
    return (2*np.pi/wavelength)*np.array([k1,k2,k3])

def get_Q(tth, eta, wavelength):
    k = get_k(tth, eta, wavelength)
    kp = get_k_prime(tth, eta, wavelength)
    return k - kp

def normal_omega(tth, eta, wavelength, omega):
    '''
    Return strained planes normal in omega coordinate system
    '''
    Qlab = get_Q(tth, eta, wavelength)
    Qomega = np.dot( np.linalg.inv( omega_matrix( omega ) ), Qlab )
    return -Qomega/np.linalg.norm(Qomega)

def taylor_strain(G_w, G_0):
    '''Strain in direction nhat by using Taylor expansion
    of the rewritten Laue equations.
    '''
    s = 1 - ( G_w.dot(G_0) /  G_0.dot(G_0) )
    return s

def strain(B, h, k, l, tth, wavelength):
    #G_original = np.sqrt((h*a_rec)**2 + (k*b_rec)**2 + (l*c_rec)**2)

    G_original = np.dot(B,np.array([h,k,l]))

    # Elm. of Modern X-ray P. pp. 157
    d_original = 2*np.pi / np.linalg.norm(G_original)

    # Bragg's law
    m = np.round(2*d_original*sind(tth/2.)/wavelength)
    d_measured = (m*wavelength)/(2*sind(tth/2.))

    return m, d_measured, d_original, (d_measured-d_original)/d_original

def weight(d_measured, d_original, strain, tth, wavelength, bragg_order, distance, pixelsize):
    '''
    Calculate weights based on resolution concerns, to be used in As=m taking
    into account that the rows of A are differently accurate measurements.

    UNITS:
          wavelength  -  arbitrary unit A
          d_original  -  arbitrary unit A
          d_measured  -  arbitrary unit A
          distance    -  arbitrary unit B
          pixelsize   -  arbitrary unit B
    '''

    tth_rad = np.radians( tth )
    r = np.tan( tth_rad )*distance # peak location radius
    dtth_rad = np.arctan( (r + pixelsize )/distance ) - tth_rad # angular width of a pixel
    d_rdr = ( (bragg_order*wavelength)/(2*np.sin( (tth_rad+dtth_rad)/2. ) ) )
    eps = (d_rdr-d_original)/d_original # strain at radius r + dr
    w = abs(1 / (strain - eps) )
    assert strain>eps
    
    return w

def uniq( vals ):
    d = {}
    newvals = []
    for v in vals:
        if v not in d:
            d[v]=0
            newvals.append(v)
    return newvals

def convert_measurements( params, grain, flt, ymin, ystep, omegastep ):

    distance = params.get('distance')
    pixelsize = ( params.get('y_size') + params.get('z_size') ) / 2.0
    wavelength = params.get('wavelength')
    cell = [params.get('cell__a'),params.get('cell__b'),params.get('cell__c'),params.get('cell_alpha'),params.get('cell_beta'),params.get('cell_gamma')]

    B_0 = tools.form_b_mat(cell)

    keys = [ (hkl[0], hkl[1], hkl[2], int(s))
             for hkl, s in zip(grain.hkl.T , grain.etasigns)]

    uni = uniq(keys)
    akeys = np.array( keys )
    strains = []
    directions = []
    all_omegas = []
    dtys = []
    all_tths = []
    all_etas = []
    weights = []
    all_intensity = []
    all_sc = []
    all_Gws = []
    all_hkl = []

    for refi,u in enumerate(uni):

        # h==h, k==k, l==l, sign==sign
        mask = (akeys == u).astype(int).sum(axis=1) == 4
        tths = flt.tth[grain.mask][mask]
        etas = flt.eta[grain.mask][mask]
        omegas = flt.omega[grain.mask][mask]
        scs = flt.sc[grain.mask][mask]
        detector_y_pos = flt.dty[ grain.mask ][mask]
        intensity = flt.sum_intensity[grain.mask][mask]
        scs = flt.sc[grain.mask][mask]
        G_ws = np.array( (flt.gx[grain.mask][mask], flt.gy[grain.mask][mask], flt.gz[grain.mask][mask]) ).T
        h = u[0]
        k = u[1]
        l = u[2]
        
        for sc, dty, tth, eta, om, I, sc, G_w in zip( scs, detector_y_pos, tths, etas, omegas, intensity, scs, G_ws):

            all_hkl.append( [h,k,l] )
            bragg_order, d_measured, d_original, eps = strain(B_0, h, k, l, tth, wavelength)

            grain.u = tools.ubi_to_u(grain.ubi)
            G_0 = grain.u.dot( B_0.dot( np.array([h,k,l]) ) )
            Qlab = get_Q(tth, eta, wavelength)
            Qomega = -np.dot( np.linalg.inv( omega_matrix( om ) ), Qlab )
            eps_taylor = taylor_strain(Qomega, G_0) #<= to use Henningssons Taylor
            
            strains.append( eps_taylor )

            #print(eps, eps_taylor)
            #strains.append( eps ) #<= to use Poulsens
            
            #directions.append( G_0/np.linalg.norm(G_0) ) #<= this approximation is probably equally good
            directions.append( normal_omega(tth, eta, wavelength, om) )
            all_omegas.append( om )
            all_tths.append( tth )
            all_etas.append( eta )
            dtys.append( dty )
            all_intensity.append( I )
            all_sc.append( sc )
            all_Gws.append(G_w)
            #plt.scatter(dty, eps)
            #plt.title(r'$\omega$ = '+str(om))
            
            weights.append( weight( d_measured, d_original, eps, tth, wavelength, bragg_order, distance, pixelsize ) )
        #plt.show()
    measurements = {}
    measurements['strain']    = np.array(strains)
    measurements['kappa']     = np.array(directions)
    measurements['omega']     = np.array(all_omegas)
    measurements['dty']       = np.array(dtys)
    measurements['sig_m']     = 1./np.array(weights)
    measurements['tth']       = np.array(all_tths)
    measurements['eta']       = np.array(all_etas)
    measurements['intensity'] = np.array(all_intensity)
    measurements['sc']        = np.array(all_sc)
    measurements['Gomega']    = np.asarray(all_Gws)
    measurements['hkl']       = np.asarray(all_hkl)

    if 0:
        sort_index = np.argsort(measurements['tth'])
        plt.plot(measurements['tth'][sort_index]/2., measurements['sig_m'][sort_index])
        plt.xlabel(r'$\theta$ [dgr]')
        plt.ylabel(r'$\sigma_m$')
        plt.grid(True)
        plt.show()
    return measurements




import torch
import numpy as np

def defineFourierBasis(m,lx,ly,lz, sig_f, nhat=None, subset=None):
    """Defines a set of Fourier basis functions used in the bayesian regression.

    Args:
        m (:obj:`torch.tensor`): number of basis functions along each axis
        lx (:obj:`torch.tensor`): lengthscale along x axis
        ly (:obj:`torch.tensor`): lengthscale along y axis
        lz (:obj:`torch.tensor`): lengthscale along z axis
        sig_f (:obj:`torch.tensor`): prior signal uncertainty
        nhat (:obj:`torch.tensor`): the unit vector directions of the measurements
    Returns: 
        (:obj:`dict`): dictionary with keys and values:

            **lambdas** (:obj:`torch.tensor`): placement of each basis function (frequencies)
  
            **SLambda** (:obj:`torch.tensor`): spectral densities for each basis function (i.e. prior variances)
  
            **Lx,Ly,Lz** (:obj:`torch.tensor`): domain size in x,y,z directions

            **Q1,Q2,Q3,Q4** (:obj:`torch.tensor`): Integral partial results to be used in :func:`basisFuncDInts`.

    """

    [mm1, mm2, mm3] = torch.meshgrid(torch.linspace(1.0,m,m),torch.linspace(1.0,m,m),torch.linspace(1.0,m,m))            # grid of basisfunctions
    insideCircle = torch.sqrt(mm1**2 + mm2**2 + mm3**2) / m <= 1 + 2e-16 # points inside a circle
    mm1 = mm1[insideCircle].double()
    mm2 = mm2[insideCircle].double()
    mm3 = mm3[insideCircle].double()

    dlambda_x = 3.5 / lx / m            # 3.5 sigmacoverage, this coudl be adjusted out to 4.5 sigma
    dlambda_y = 3.5 / ly / m
    dlambda_z = 3.5 / lz / m

    lambda_x = mm1.unsqueeze(0)*dlambda_x        # lambdas
    lambda_y = mm2.unsqueeze(0)*dlambda_y
    lambda_z = mm3.unsqueeze(0)*dlambda_z
    
    Lx = np.pi / 2 / dlambda_x          # Domain scales
    Ly = np.pi / 2 / dlambda_y
    Lz = np.pi / 2 / dlambda_z

    Basis = {'Lx': Lx,
             'Ly': Ly,
             'Lz': Lz}

    if nhat is not None:
        n1 = nhat[[0], :].T
        n2 = nhat[[1], :].T
        n3 = nhat[[2], :].T

        # could calculate this bit in the define basis function part
        n1lx = n1 * lambda_x
        n2ly = n2 * lambda_y
        n3lz = n3 * lambda_z
        Q1 = (n1lx - n2ly - n3lz)
        Q2 = (n1lx + n2ly - n3lz)
        Q3 = (n1lx - n2ly + n3lz)
        Q4 = (n1lx + n2ly + n3lz)

        # look for zeros that will later cause divide by zero errors
        I1 = (Q1.abs() < 1e-10).any(0)
        I2 = (Q2.abs() < 1e-10).any(0)
        I3 = (Q3.abs() < 1e-10).any(0)
        I4 = (Q4.abs() < 1e-10).any(0)

        singular_index = (I1+I2+I3+I4).detach().numpy().astype(np.int)
        max_lam_x = lambda_x.max().detach().numpy()
        max_lam_y = lambda_y.max().detach().numpy()
        max_lam_z = lambda_z.max().detach().numpy()

        # if max_lam is very small the while() will need super many iterations,
        # if max_lam==0 it will fail completely. Best to set a lower bound for
        # these scaling parameters
        if np.abs(max_lam_x)<1.5e-4: max_lam_x = 1.5e-4
        if np.abs(max_lam_y)<1.5e-4: max_lam_y = 1.5e-4
        if np.abs(max_lam_z)<1.5e-4: max_lam_z = 1.5e-4

        maxiteration = 500
        iteration = 0
        while(singular_index.any() and iteration<=maxiteration):
            iteration+=1
            mask = torch.tensor(singular_index)

            lambda_x = lambda_x + mask * 1.5e-6 * max_lam_x * torch.rand(lambda_x.shape)
            lambda_y = lambda_y + mask * 1.5e-6 * max_lam_y * torch.rand(lambda_y.shape)
            lambda_z = lambda_z + mask * 1.5e-6 * max_lam_z * torch.rand(lambda_z.shape)

            n1lx = n1 * lambda_x
            n2ly = n2 * lambda_y
            n3lz = n3 * lambda_z

            Q1 = (n1lx - n2ly - n3lz)
            Q2 = (n1lx + n2ly - n3lz)
            Q3 = (n1lx - n2ly + n3lz)
            Q4 = (n1lx + n2ly + n3lz)

            # look for zeros that will later cause divide by zero errors
            I1 = (Q1.abs() < 1e-10).any(0)
            I2 = (Q2.abs() < 1e-10).any(0)
            I3 = (Q3.abs() < 1e-10).any(0)
            I4 = (Q4.abs() < 1e-10).any(0)
            singular_index = (I1+I2+I3+I4).detach().numpy().astype(np.int)
        assert iteration<maxiteration, 'defineFourierBasis() was unable to remove numerical instability'

        Basis['Q1'] = Q1
        Basis['Q2'] = Q2
        Basis['Q3'] = Q3
        Basis['Q4'] = Q4

    lambdas = torch.cat([lambda_x, lambda_y, lambda_z], dim=0)

    SLambda = sig_f**2 * (2 * np.pi)**(3 / 2) * lx * ly * lz * torch.exp(
        -0.5 * (lambda_x**2 * lx**2 + lambda_y**2 * ly**2 + lambda_z**2 * lz**2))

    Basis['lambdas'] = lambdas
    Basis['SLambda'] = SLambda

    return Basis

def defineBeltramiBasis(m, theta, nhat=None):
    """Defines a set of Fourier basis functions used in the bayesian regression.

    Args:
        m (:obj:`torch.tensor`): Number of basis functions in each frequency dimension.
        theta (:obj:`torch.tensor`): Hyper parameters defining covariance function.
        nhat (:obj:`torch.tensor`): X-ray beam directionin sample coordinate system.

    Returns: 
        (:obj:`tuple` of :obj:`dict`): tuple of dictionaries dictionaries of type :func:`defineFourierBasis`.
        one dictionary for each beltrami tensor componenent. Also returns a concatenated tensor with all 
        dictionaries contained.

    """

    A = defineFourierBasis(m, theta[1], theta[2], theta[3], theta[0], nhat)
    B = defineFourierBasis(m, theta[5], theta[6], theta[7], theta[4], nhat)
    C = defineFourierBasis(m, theta[9], theta[10], theta[11], theta[8], nhat)
    D = defineFourierBasis(m, theta[13], theta[14], theta[15], theta[12], nhat)
    E = defineFourierBasis(m, theta[17], theta[18], theta[19], theta[16], nhat)
    F = defineFourierBasis(m, theta[21], theta[22], theta[23], theta[20], nhat)

    SLambda = torch.cat([A['SLambda'], B['SLambda'], C['SLambda'], D['SLambda'], E['SLambda'], F['SLambda']], 1)

    return A,B,C,D,E,F,SLambda

def basisFuncDerivs(lambdas, Lx, Ly, Lz, x, y, z, req):
    """ Calculates the basis function derivatives Beltrami basis components.

    Args:
        lambdas (:obj:`torch.tensor`): Basis function frequencies.
        Lx,Ly,Lz (:obj:`torch.tensor`): Basis function phases in x,y,z domain.
        x,y,z (:obj:`torch.tensor`): Coordinates at which to calculate basis functions at.
        req (:obj:`list` of :obj:`bool`): Components desired to calculate for.

    Returns: 
        (:obj:`torch.tensor` of :obj:`dict`): Partial derivatives of the basis function, ```shape=6,n,m)```.

    """
    n = x.shape[0]                         # number of points
    mm_adj = lambdas.shape[1]               # total number of basis functions

    # preallocate output
    dfuncs = torch.zeros((n,mm_adj,6)).double()

    # point/measurement stuff always indexes along rows,
    # basis function stuff always indexes along columns,
    Bx = (x + Lx).unsqueeze(1) * lambdas[[0], :]
    By = (y + Ly).unsqueeze(1) * lambdas[[1], :]
    Bz = (z + Lz).unsqueeze(1) * lambdas[[2], :]

    sx = torch.sin(Bx)
    sy = torch.sin(By)
    sz = torch.sin(Bz)
    cx = torch.cos(Bx)
    cy = torch.cos(By)
    cz = torch.cos(Bz)

    lambda_x = lambdas[[0],:]
    lambda_y = lambdas[[1],:]
    lambda_z = lambdas[[2],:]

    if req[0]:   # d/dxdx
        dfuncs[:,:,0] = -(lambda_x**2/torch.sqrt(Lx*Ly*Lz)) * sx * sy * sz

    if req[1]:   # d/dydy
        dfuncs[:,:,1] = -(lambda_y**2/torch.sqrt(Lx*Ly*Lz)) * sx * sy* sz

    if req[2]:   # d/dzdz
        dfuncs[:,:,2] = -(lambda_z**2/torch.sqrt(Lx*Ly*Lz)) * sx * sy * sz

    if req[3]:   # d/dxdy
        dfuncs[:,:,3] = ((lambda_x*lambda_y)/torch.sqrt(Lx*Ly*Lz)) * cx * cy * sz

    if req[4]:   # d/dxdz
        dfuncs[:,:,4] = ((lambda_x*lambda_z)/torch.sqrt(Lx*Ly*Lz)) * cx * sy * cz

    if req[5]:   # d/dydz
        dfuncs[:,:,5] = ((lambda_y*lambda_z)/torch.sqrt(Lx*Ly*Lz)) * sx * cy * cz

    return dfuncs

def BeltramiApproxAnisoStrainField(A,B,C,D,E,F,x,y,z,H):
    """Calculates regressor matrix given hyperparameters for beltrami components and spatial locations.

    This function works for aniostropic materials as the compliance H can be supplied.

    Args:
        A (:obj:`dict`): Basis fuctions for Beltrami component A
        B (:obj:`dict`): Basis fuctions for Beltrami component B
        C (:obj:`dict`): Basis fuctions for Beltrami component C
        D (:obj:`dict`): Basis fuctions for Beltrami component D
        E (:obj:`dict`): Basis fuctions for Beltrami component E
        F (:obj:`dict`): Basis fuctions for Beltrami component F
        x (:obj:`torch.tensor`): x coordinate of points to calculate basis functions at
        y (:obj:`torch.tensor`): y coordinate of points to calculate basis functions at
        z (:obj:`torch.tensor`): z coordinate of points to calculate basis functions at
        H (:obj:`torch.tensor`): the compliance matrix, either a ```shape=(6,6)``` average 
            for the material or ```shape=(N,6,6)``` individual compliances for each point

    Returns: 
        The regressor matrix

    """

    n = len(x)                      # number of test/data points
    mm_adj = A['lambdas'].shape[1]  # total number of basis functions per component

    dfuncs = basisFuncDerivs(A['lambdas'],A['Lx'],A['Ly'],A['Lz'], x, y, z,[0, 1, 1, 0, 0, 1])
    dydyA = dfuncs[:,:,1]
    dzdzA = dfuncs[:,:,2]
    dydzA = dfuncs[:,:,5]

    dfuncs = basisFuncDerivs(B['lambdas'],B['Lx'],B['Ly'],B['Lz'], x, y, z,[1, 0, 1, 0, 1, 0])
    dxdxB = dfuncs[:,:,0]
    dzdzB = dfuncs[:,:,2]
    dxdzB = dfuncs[:,:,4]

    dfuncs = basisFuncDerivs(C['lambdas'],C['Lx'],C['Ly'],C['Lz'], x, y, z,[1, 1, 0, 1, 0, 0])
    dxdxC = dfuncs[:,:,0]
    dydyC = dfuncs[:,:,1]
    dxdyC = dfuncs[:,:,3]

    dfuncs = basisFuncDerivs(D['lambdas'],D['Lx'],D['Ly'],D['Lz'], x, y, z,[0, 0, 1, 1, 1, 1])
    dzdzD = dfuncs[:,:,2]
    dxdyD = dfuncs[:,:,3]
    dxdzD = dfuncs[:,:,4]
    dydzD = dfuncs[:,:,5]

    dfuncs = basisFuncDerivs(E['lambdas'],E['Lx'],E['Ly'],E['Lz'], x, y, z,[0, 1, 0, 1, 1, 1])
    dydyE = dfuncs[:,:,1]
    dxdyE = dfuncs[:,:,3]
    dxdzE = dfuncs[:,:,4]
    dydzE = dfuncs[:,:,5]

    dfuncs = basisFuncDerivs(F['lambdas'],F['Lx'],F['Ly'],F['Lz'], x, y, z,[1, 0, 0, 1, 1, 1])
    dxdxF = dfuncs[:,:,0]
    dxdyF = dfuncs[:,:,3]
    dxdzF = dfuncs[:,:,4]
    dydzF = dfuncs[:,:,5]

    # the basis functions for stress
    phi_sxx = torch.cat([torch.zeros((n, mm_adj)), dzdzB, dydyC, torch.zeros(n, mm_adj), torch.zeros((n, mm_adj)), -2 * dydzF],1)
    phi_syy = torch.cat([dzdzA, torch.zeros((n, mm_adj)), dxdxC, torch.zeros((n, mm_adj)), -2 * dxdzE, torch.zeros((n, mm_adj))],1)
    phi_szz = torch.cat([dydyA, dxdxB, torch.zeros((n, mm_adj)), -2 * dxdyD, torch.zeros((n, mm_adj)), torch.zeros((n, mm_adj))],1)
    phi_sxy = torch.cat([torch.zeros((n, mm_adj)), torch.zeros((n, mm_adj)), -dxdyC, -dzdzD, dydzE, dxdzF],1)
    phi_sxz = torch.cat([torch.zeros((n, mm_adj)), -dxdzB, torch.zeros((n, mm_adj)), dxdzD, dxdyE, -dxdxF],1)
    phi_syz = torch.cat([-dydzA, torch.zeros((n, mm_adj)), torch.zeros((n, mm_adj)), dydzD, -dydyE, dxdyF],1)

    # combined and reshaped to be size [n,6,6*mm_adj]
    phi_s = torch.cat([phi_sxx.unsqueeze(1),phi_syy.unsqueeze(1),phi_szz.unsqueeze(1),phi_sxy.unsqueeze(1),phi_sxz.unsqueeze(1),phi_syz.unsqueeze(1)],1)

    # apply hooke's law to get the basis functions for strain as a [n,6,6*mm_adj]
    phi = torch.matmul(H, phi_s)
    
    # and flatten to be [6*N,6*mm_adj]
    return phi.reshape((6*n,6*mm_adj))

def BeltramiApproxAnisoStressField(A,B,C,D,E,F,x,y,z):
    """Calculates regressor matrix given hyperparameters for beltrami components and spatial locations.

    This function works for aniostropic materials as the compliance H can be supplied.

    Args:
        A (:obj:`dict`): Basis fuctions for Beltrami component A
        B (:obj:`dict`): Basis fuctions for Beltrami component B
        C (:obj:`dict`): Basis fuctions for Beltrami component C
        D (:obj:`dict`): Basis fuctions for Beltrami component D
        E (:obj:`dict`): Basis fuctions for Beltrami component E
        F (:obj:`dict`): Basis fuctions for Beltrami component F
        x (:obj:`torch.tensor`): x coordinate of points to calculate basis functions at
        y (:obj:`torch.tensor`): y coordinate of points to calculate basis functions at
        z (:obj:`torch.tensor`): z coordinate of points to calculate basis functions at

    Returns: 
        The regressor matrix

    NOTE: Approximating stress is simply done by approximating strain with H=I unity.
    """
    I = torch.tensor(np.expand_dims( np.eye(6,6), axis=0)).double()
    return BeltramiApproxAnisoStrainField(A,B,C,D,E,F,x,y,z,H=I)

def basisFuncDInts(lambdas, Lx, Ly, Lz, entry, exit, req, nsegs, Q1, Q2, Q3, Q4):
    """Computes line integrals of the derivatives of basis functions.

    Args:
        lambdas (:obj:`torch.tensor`): placement of each basis function (frequencies)
        Lx,Ly,Lz (:obj:`torch.tensor`): domain size in x,y,z directions
        entry (:obj:`torch.tensor`): entry points of each segment of each measurement ray
        exit (:obj:`torch.tensor`): exit points of each segment of each measurement ray
        req (:obj:`list` of :obj:`bool`): Components desired to calculate for.
        nsegs (:obj:`torch.tensor`): Number of segments for each line integral. ``shape=(N,)``
        Q1,Q2,Q3,Q4 (:obj:`torch.tensor`): Integral partial results computed by :func:`defineFourierBasis`.

    Returns: 
        :obj:`torch.tensor` of ```shape=(n, mm_adj, 6)```, the line integrals of the derivatives of each basis function

    """
    n = entry.shape[1]                      # number of LRTs
    mm_adj = lambdas.shape[1]               # total number of basis functions

    lambda_x = lambdas[[0],:]   # maintain these as row vectors
    lambda_y = lambdas[[1],:]
    lambda_z = lambdas[[2],:]

    # preallocate
    Idfuncs = torch.zeros((n, mm_adj, 6)).double()
    for ss in range(int(nsegs.max().item())):
        segInd = (nsegs > ss).squeeze()      # index of measurements with at least ss+1 segments

        # entry coordinates for this ray segment for each measurement with at least this many segments
        x0 = entry[ss*3,segInd].unsqueeze(1)
        y0 = entry[ss*3+1,segInd].unsqueeze(1)
        z0 = entry[ss*3+2,segInd].unsqueeze(1) # make these column vectors

        alpha_x = (Lx + x0) * lambda_x
        alpha_y = (Ly + y0) * lambda_y
        alpha_z = (Lz + z0) * lambda_z

        Gamma1s = torch.cos(alpha_x - alpha_y - alpha_z)
        Gamma2s = torch.cos(alpha_x + alpha_y - alpha_z)
        Gamma3s = torch.cos(alpha_x - alpha_y + alpha_z)
        Gamma4s = torch.cos(alpha_x + alpha_y + alpha_z)

        # exit coordinates for this ray segment for each measurement with at least this many segments
        xf = exit[ss*3, segInd].unsqueeze(1)
        yf = exit[ss*3+1, segInd].unsqueeze(1)
        zf = exit[ss *3+2, segInd].unsqueeze(1)  # make these column vectors

        alpha_x = (Lx + xf) * lambda_x
        alpha_y = (Ly + yf) * lambda_y
        alpha_z = (Lz + zf) * lambda_z

        Gamma1f = torch.cos(alpha_x - alpha_y - alpha_z)
        Gamma2f = torch.cos(alpha_x + alpha_y - alpha_z)
        Gamma3f = torch.cos(alpha_x - alpha_y + alpha_z)
        Gamma4f = torch.cos(alpha_x + alpha_y + alpha_z)

        Gamma1 = (Gamma1f - Gamma1s) / Q1[segInd,:]    # TODO: how to protect this form having hte wrong shape if segInd is scalar?
        Gamma2 = (Gamma2f - Gamma2s) / Q2[segInd,:]
        Gamma3 = (Gamma3f - Gamma3s) / Q3[segInd,:]
        Gamma4 = (Gamma4f - Gamma4s) / Q4[segInd,:]

        if req[0]: # integral of d / dxdx
            Idfuncs[segInd,:, 0] = Idfuncs[segInd,:, 0] - (lambda_x ** 2 / torch.sqrt(Lx * Ly * Lz) / 4) * (
                    Gamma1 - Gamma2 - Gamma3 + Gamma4)

        if req[1]: # integral of d / dydy
            Idfuncs[segInd,:, 1] = Idfuncs[segInd,:, 1] - (lambda_y ** 2 / torch.sqrt(Lx * Ly * Lz) / 4) * (
                    Gamma1 - Gamma2 - Gamma3 + Gamma4)

        if req[2]: # integral of d / dzdz
            Idfuncs[segInd,:, 2] = Idfuncs[segInd,:, 2] -(lambda_z ** 2 / torch.sqrt(Lx * Ly * Lz) / 4) * (Gamma1-Gamma2-Gamma3+Gamma4)

        if req[3]: # integral of d / dxdy
            Idfuncs[segInd,:, 3] = Idfuncs[segInd,:, 3] +(lambda_x * lambda_y / torch.sqrt(Lx * Ly * Lz) / 4) * (
                    Gamma1 + Gamma2 - Gamma3 - Gamma4)

        if req[4]: # integral of d / dxdz
            Idfuncs[segInd,:, 4] = Idfuncs[segInd,:, 4] +(lambda_x * lambda_z / torch.sqrt(Lx * Ly * Lz) / 4) * (
                    Gamma1 - Gamma2 + Gamma3 - Gamma4)

        if req[5]: # integral of d / dydz
            Idfuncs[segInd,:, 5] = Idfuncs[segInd,:, 5] +(lambda_y * lambda_z / torch.sqrt(Lx * Ly * Lz) / 4) * (
                    -Gamma1 - Gamma2 - Gamma3 - Gamma4)
    return Idfuncs

def BeltramiApproxAnisoRayMeas(A,B,C,D,E,F,H,entry,exit,nsegs,L, kappa,subset_idx=None):
    """
    """
    n = entry.shape[1]                      # number of ray measurements
    mm_adj = A['lambdas'].shape[1]  # total number of basis functions per component
    idx = subset_idx

    if idx is None:
        Idfuncs = basisFuncDInts(A['lambdas'], A['Lx'], A['Ly'], A['Lz'], entry, exit, [0, 1, 1, 0, 0, 1], nsegs,
                       A['Q1'], A['Q2'], A['Q3'], A['Q4'])
    else:
        Idfuncs = basisFuncDInts(A['lambdas'], A['Lx'], A['Ly'], A['Lz'], entry, exit, [0, 1, 1, 0, 0, 1], nsegs,
                       A['Q1'][idx,:], A['Q2'][idx,:], A['Q3'][idx,:], A['Q4'][idx,:])
    IdydyA = Idfuncs[:,:,1]
    IdzdzA = Idfuncs[:,:,2]
    IdydzA = Idfuncs[:,:,5]

    if idx is None:
        Idfuncs = basisFuncDInts(B['lambdas'], B['Lx'], B['Ly'], B['Lz'], entry, exit, [1, 0, 1, 0, 1, 0], nsegs,
                   B['Q1'], B['Q2'], B['Q3'], B['Q4'])
    else:
        Idfuncs = basisFuncDInts(B['lambdas'], B['Lx'], B['Ly'], B['Lz'], entry, exit, [1, 0, 1, 0, 1, 0], nsegs,
                   B['Q1'][idx,:], B['Q2'][idx,:], B['Q3'][idx,:], B['Q4'][idx,:])
    IdxdxB = Idfuncs[:,:,0]
    IdzdzB = Idfuncs[:,:,2]
    IdxdzB = Idfuncs[:,:,4]

    if idx is None:
        Idfuncs = basisFuncDInts(C['lambdas'], C['Lx'], C['Ly'], C['Lz'], entry, exit, [1, 1, 0, 1, 0, 0], nsegs,
                   C['Q1'], C['Q2'], C['Q3'], C['Q4'])
    else:
        Idfuncs = basisFuncDInts(C['lambdas'], C['Lx'], C['Ly'], C['Lz'], entry, exit, [1, 1, 0, 1, 0, 0], nsegs,
                   C['Q1'][idx,:], C['Q2'][idx,:], C['Q3'][idx,:], C['Q4'][idx,:])
    IdxdxC = Idfuncs[:,:,0]
    IdydyC = Idfuncs[:,:,1]
    IdxdyC = Idfuncs[:,:,3]

    if idx is None:
        Idfuncs = basisFuncDInts(D['lambdas'], D['Lx'], D['Ly'], D['Lz'], entry, exit, [0, 0, 1, 1, 1, 1], nsegs,
                   D['Q1'], D['Q2'], D['Q3'], D['Q4'])
    else:
        Idfuncs = basisFuncDInts(D['lambdas'], D['Lx'], D['Ly'], D['Lz'], entry, exit, [0, 0, 1, 1, 1, 1], nsegs,
                   D['Q1'][idx,:], D['Q2'][idx,:], D['Q3'][idx,:], D['Q4'][idx,:])
    IdzdzD = Idfuncs[:,:,2]
    IdxdyD = Idfuncs[:,:,3]
    IdxdzD = Idfuncs[:,:,4]
    IdydzD = Idfuncs[:,:,5]

    if idx is None:
        Idfuncs = basisFuncDInts(E['lambdas'], E['Lx'], E['Ly'], E['Lz'], entry, exit, [0, 1, 0, 1, 1, 1], nsegs,
                   E['Q1'], E['Q2'], E['Q3'], E['Q4'])
    else:
        Idfuncs = basisFuncDInts(E['lambdas'], E['Lx'], E['Ly'], E['Lz'], entry, exit, [0, 1, 0, 1, 1, 1], nsegs,
                   E['Q1'][idx,:], E['Q2'][idx,:], E['Q3'][idx,:], E['Q4'][idx,:])
    IdydyE = Idfuncs[:,:,1]
    IdxdyE = Idfuncs[:,:,3]
    IdxdzE = Idfuncs[:,:,4]
    IdydzE = Idfuncs[:,:,5]

    if idx is None:
        Idfuncs = basisFuncDInts(F['lambdas'], F['Lx'], F['Ly'], F['Lz'], entry, exit, [1, 0, 0, 1, 1, 1], nsegs,
                   F['Q1'], F['Q2'], F['Q3'], F['Q4'])
    else:
        Idfuncs = basisFuncDInts(F['lambdas'], F['Lx'], F['Ly'], F['Lz'], entry, exit, [1, 0, 0, 1, 1, 1], nsegs,
                   F['Q1'][idx,:], F['Q2'][idx,:], F['Q3'][idx,:], F['Q4'][idx,:])
    IdxdxF = Idfuncs[:,:,0]
    IdxdyF = Idfuncs[:,:,3]
    IdxdzF = Idfuncs[:,:,4]
    IdydzF = Idfuncs[:,:,5]

    K1K1 = kappa[[0],:].T**2    # ensuring things to do with measurements are column vectors
    K2K2 = kappa[[1],:].T**2
    K3K3 = kappa[[2],:].T**2
    K1K2 = 2 * kappa[[0],:].T * kappa[[1],:].T
    K1K3 = 2 * kappa[[0], :].T * kappa[[2], :].T
    K2K3 = 2 * kappa[[1], :].T * kappa[[2], :].T

    # integrals of the basis functions for stress
    phi_Isxx = torch.cat([torch.zeros((n, mm_adj)), IdzdzB, IdydyC, torch.zeros(n, mm_adj), torch.zeros((n, mm_adj)), -2 * IdydzF],1)
    phi_Isyy = torch.cat([IdzdzA, torch.zeros((n, mm_adj)), IdxdxC, torch.zeros((n, mm_adj)), -2 * IdxdzE, torch.zeros((n, mm_adj))],1)
    phi_Iszz = torch.cat([IdydyA, IdxdxB, torch.zeros((n, mm_adj)), -2 * IdxdyD, torch.zeros((n, mm_adj)), torch.zeros((n, mm_adj))],1)
    phi_Isxy = torch.cat([torch.zeros((n, mm_adj)), torch.zeros((n, mm_adj)), -IdxdyC, -IdzdzD, IdydzE, IdxdzF],1)
    phi_Isxz = torch.cat([torch.zeros((n, mm_adj)), -IdxdzB, torch.zeros((n, mm_adj)), IdxdzD, IdxdyE, -IdxdxF],1)
    phi_Isyz = torch.cat([-IdydzA, torch.zeros((n, mm_adj)), torch.zeros((n, mm_adj)), IdydzD, -IdydyE, IdxdyF],1)

    # combined and reshaped to be size [n,6,6*mm_adj]
    phi_Is = torch.cat([phi_Isxx.unsqueeze(1),phi_Isyy.unsqueeze(1),phi_Iszz.unsqueeze(1),phi_Isxy.unsqueeze(1),
                        phi_Isxz.unsqueeze(1),phi_Isyz.unsqueeze(1)],1)

    # converted to integrals of the strain basis functions by applying hooke's law
    phi_Istrain = torch.matmul(H,phi_Is)

    # converted to the basis function of the measurements by applying \bar{\kappa}
    Phi = K1K1 * phi_Istrain[:,0,:] + K2K2 * phi_Istrain[:,1,:] + K3K3 * phi_Istrain[:,2,:] + \
          K1K2 * phi_Istrain[:,3,:] + K1K3 * phi_Istrain[:,4,:] + K2K3 * phi_Istrain[:,5,:]

    return Phi / L.view(n, 1)  # make sure L is column vec

def negativeLogLikelihood(Phi, Y, sig_m, v, r):
    """Computes the negative log likelihood based on gaussian measurement model.

    cost =  0.5*error^T Sigma^{-1} error + 0.5 log(det(Sigma)) + C
    where Sigma is the predictions of the measurement variances
    Sigma = Variance of the predicted value + variance of the measurements (sig_m^2)
    This computation is done in square root form for computational efficiency and
    numerical accuracy

    Args:
        Phi (:obj:`torch.tensor`): the regressor matrix
        Y (:obj:`torch.tensor`): the measurement values
        sig_m (:obj:`torch.tensor`): the measurement standard deviations, either a scalar or an N,1 vector of individual stds
        v (:obj:`torch.tensor`): the fit basis function coefficients
        r (:obj:`torch.tensor`): the cholesky factor calculated during regression

    Returns:
        The negative log likelihood

    """
    predictions, chol_var = predict(Phi, v, r)
    n = len(predictions)
    Gamma = torch.cat([chol_var, torch.diag(sig_m.squeeze()*torch.ones(n,))], 0)
    (_, r) = torch.qr(Gamma)
    error = Y - predictions
    (alpha, _) = torch.triangular_solve(error.unsqueeze(1), r, transpose=True)
    nll = 0.5 * torch.dot(alpha.squeeze(),alpha.squeeze()) + r.diag().abs().log().sum()
    return nll


def nll_scipy_wrapper_RT_aniso(logtheta, Y_train, entry_train, exit_train, nsegs_train, L_train, kappa_train,
                          Y_test, entry_test,exit_test,nsegs_test,L_test, kappa_test, sig_m_train, sig_m_test,
                         Hray_train, Hray_test, m, nhat, train_idx, test_idx):
    """Computes the negative log likelihood based on gaussian measurement model.
    """
    logtheta = torch.tensor(logtheta, requires_grad=True).double()
    theta = logtheta.exp()
    A, B, C, D, E, F, SLambda = defineBeltramiBasis(m, theta[:24], nhat=nhat)
    Phi_train = BeltramiApproxAnisoRayMeas(A, B, C, D, E, F, Hray_train, entry_train,
                                      exit_train, nsegs_train, L_train, kappa_train, subset_idx=train_idx)
    v, r = regression(Phi_train, Y_train, sig_m_train, SLambda)
    Phi_test = BeltramiApproxAnisoRayMeas(A, B, C, D, E, F, Hray_test, entry_test,
                                     exit_test, nsegs_test, L_test, kappa_test, subset_idx=test_idx)
    loss = negativeLogLikelihood(Phi_test, Y_test, sig_m_test, v, r)
    loss.backward()
    grad = logtheta.grad
    return loss.detach().numpy(), grad.detach().numpy()

def regression(Phi, Y, sig_m, SLambda):
    """Performs linear bayesian regression to fit basis function coefficients.

    Args:
        Phi (:obj:`torch.tensor`): The regressor matrix
        Y (:obj:`torch.tensor`): The measurements
        sig_m (:obj:`torch.tensor`): The measurement standard deviation, either a scalar or a N,1 array of individual stds
        SLambda (:obj:`torch.tensor`): The prior covariances for the basis functions

    Returns:
        (v,r) where v is the basis function coefficients and r is the cholesky factor required form stuffs

    """
    # n, m = Phi.shape
    Gamma = torch.cat([Phi / sig_m, torch.diag(1 / SLambda.sqrt().squeeze())], 0)
    (_, r) = torch.qr(Gamma)  # does a thing QR decomp
    (tmp, _) = torch.triangular_solve(Phi.T.matmul(Y.unsqueeze(1) / sig_m ** 2), r, transpose=True)
    (v, _) = torch.triangular_solve(tmp, r, transpose=False)
    return v, r

def predict(Phi, v, r):
    """Predicts the point location values of a field given regressor matrix.
    
    Based on the fitted basis function coefficients v, and the cholesky factor r
    also returns the cholesky factor of the variances of these predictions

    Args:
        Phi (:obj:`torch.tensor`): The regressor matrix
        v (:obj:`torch.tensor`): The fit basis function coefficients
        r (:obj:`torch.tensor`): The cholesky factor calculated during regression

    Returns:
        predictions, chol_var
    """
    predictions = Phi.matmul(v).squeeze()
    (chol_var, _) = torch.triangular_solve(Phi.T, r, transpose=True)
    return predictions, chol_var


import numpy as np
from scipy.integrate import quad
import matplotlib.pyplot as plt
from . import analytical_strain_fields as asf

def project_strain( strain, Xgrid, Ygrid, Zgrid, entry, exit, nhat , kappa, L, nsegs, nfirst ):
    '''Forward model average strain along ray paths given an analytical
    strain tensor field and some ray geometries.

    Input:
        strain : 
        entry  : is a 3 by N array of the coordinates where the ray enters the sample (each col is x, y, z)
        exit   : is a 3 by N array of the coordinates where the ray leaves the sample (each col is x, y,z)
        nhat   : is a 3 by N array of directions of each ray (each col is a unit vector)
        kappa  : is a 3 by N array that defines the direction of strain measured (each col is a unit vector)
        L      : is a 1 by N array of the total irradiated lengths of each ray (ie distance between entry and exit)
        nsegs  : is a 1 by N array of number of segments for each ray
    Returns:
        Y      : is N by 1 array of forward modeled measurement values
    '''
    Y = np.zeros( (L.shape[1],) )

    for i in range( nfirst ):

        y = 0
        indx = np.argmin( np.abs( Zgrid.flatten()-entry[2, i] ) )
        mask = ( Zgrid.flatten()==Zgrid.flatten()[indx] )
        xvec = Xgrid.flatten()[mask]
        yvec = Ygrid.flatten()[mask]

        eps = [ s.flatten()[mask] for s in strain]
        strain_tens_func = asf.discrete_to_analytic( eps, xvec, yvec )

        for j in range( int(nsegs[0,i]) ):

            p1 = entry[3*j:3*j + 3,i] 
            p2 = exit[3*j:3*j + 3,i]
            l = np.linalg.norm(p2-p1)

            def func(s):
                epsilon = strain_tens_func(p1 + s*nhat[:,i])
                return kappa[:,i].T.dot( epsilon ).dot( kappa[:,i] )

            y += quad(func, 0, l)[0]

        Y[i] = y/L[0,i]
    
    return Y

        


import numpy as np
from scipy.optimize import leastsq
from ImageD11 import parameters, grain, transform
import copy
from . import scanning_transform
from skimage.transform import warp
import matplotlib.pyplot as plt

def fit_one_grain( gr, flt, pars):
    """
    Uses scipy.optimize to fit a single grain
    """
    args = flt, pars, gr
    ub = np.linalg.inv(gr.ubi)
    x0 = ub.ravel().copy()
    xf, cov_v, info, mesg, ier = leastsq(
        calc_teo_fit, x0, args, full_output=True)
    ub = xf.copy()
    ub.shape = 3, 3
    ubi = np.linalg.inv(ub)

    gr.set_ubi(ubi)

def calc_teo_fit( ub, flt, pars, gr, return_concatenated=True):
    """
    Function for refining ub using tth, eta, omega data
    ub is the parameter array to fit
    flt is all the data
    pars in the diffractometer geometry to get tthcalc, etacalc, omegacalc
    gr is the grain holding the peak assignments
    flt.wtth, weta, wometa = weighting functions for tth vs eta vs omega errors
    """
    UB = np.array(ub)
    UB.shape = 3, 3

    tthcalc, etacalc, omegacalc = calc_tth_eta_omega(
        UB, gr.hkl, pars, gr.etasigns)

    dtth = (flt.tth[gr.mask] - tthcalc) * flt.wtth[gr.mask]
    deta = (flt.eta[gr.mask] - etacalc) * flt.weta[gr.mask]
    domega = (flt.omega[gr.mask] - omegacalc) * flt.womega[gr.mask]
    if return_concatenated:
        return np.concatenate((dtth, deta, domega))
    else:
        return dtth, deta, domega

def calc_tth_eta_omega( ub, hkls, pars, etasigns):
    """
    Predict the tth, eta, omega for each grain
    ub = ub matrix (inverse ubi)
    hkls = peaks to predict
    pars = diffractometer info (wavelength, rotation axis)
    etasigns = which solution for omega/eta to choose (+y or -y)
    """
    g = np.dot(ub, hkls)

    tthcalc, eta2, omega2 = transform.uncompute_g_vectors(g,  pars.get('wavelength'),
                                                            wedge=pars.get('wedge'),
                                                            chi=pars.get('chi'))
    # choose which solution (eta+ or eta-)
    e0 = np.sign(eta2[0]) == etasigns
    etacalc = np.where(e0, eta2[0], eta2[1])
    omegacalc = np.where(e0, omega2[0], omega2[1])
    return tthcalc, etacalc, omegacalc

def get_peak_quantities( flt, pars, grain=None ):
    """
    Compute twotheta, eta, g-vector for given grain. Note that the entire peak
    set will be placed at the grain cms, no respect is taken to the grian peak mask.
    if grain is None, it is assumed that all scattering occurs from the lab origin.
    """
    if grain is None:
        pars.parameters['t_x'] = np.zeros(flt.nrows)
        pars.parameters['t_y'] = np.zeros(flt.nrows)
        pars.parameters['t_z'] = np.zeros(flt.nrows)
    else:
        pars.parameters['t_x'] = get_cms_along_beam( grain.recon, 
                                                    flt.omega, 
                                                    flt.dty, 
                                                    pars.parameters['ymin'], 
                                                    pars.parameters['ystep'] )       
        pars.parameters['t_y'] = np.zeros(flt.nrows)
        pars.parameters['t_z'] = np.zeros(flt.nrows)

    tth, eta = scanning_transform.compute_tth_eta( [flt.sc, flt.fc], omega=flt.omega, **pars.parameters)

    gve = scanning_transform.compute_g_vectors(tth, eta, flt.omega,
                                        pars.get('wavelength'),
                                        wedge=pars.get('wedge'),
                                        chi=pars.get('chi'))

    pars.parameters['t_x'] = None
    pars.parameters['t_y'] = None
    pars.parameters['t_z'] = None

    return tth, eta, gve

def estimate_weights( pars, flt, OMSLOP, g=None):
    distance = pars.get('distance')
    pixelsize = (pars.get('y_size') + pars.get('z_size')) / 2.0
    # 1 pixel - high energy far detector approximation
    if g:
        wtth = np.ones(np.sum(g.mask)) / np.degrees(pixelsize / distance)
        weta = wtth * np.tan(np.radians(flt.tth[g.mask]))
        womega = np.ones(np.sum(g.mask))/OMSLOP
    else:
        wtth = np.ones(flt.nrows) / np.degrees(pixelsize / distance)
        weta = wtth * np.tan(np.radians(flt.tth))
        womega = np.ones(flt.nrows)/OMSLOP

    return wtth, weta, womega


def get_cms_along_beam( recon, omega, dty, ymin, ystep, stepsize=0.5 ):
    """Compute grain centroid component along lab - x axis (beam direction) based on a 2d pixel reconstruction.

    This function rotates the reconstructed grain shapes by provided omega angles and computes the centroid 
    along the beam. To speed up computation, only every stepsize degrees is computed for, and the input angles
    are mapped to closest angle.

    Args:
        recon (:obj:`numpy array`): Pixelated reconstruction of grain shape, ```shape=(m,m)```.
        omega (:obj:`numpy array`): Projection angles in units of degrees, ```shape=(n,)```.
        ystep (:obj:`float`): Pixel size of reconstruction
        stepsize (:obj:`float`): degree stepsize resolution only these rotations are used (for speed).

    Returns:
        ```cms_along_beam``` 1d ```numpy array``` with centroid x coordinates, ```shape=(n,)```.

    """

    assert recon.shape[0]==recon.shape[1], "Only square recons will work"
    
    cms_weights = np.linspace(0,recon.shape[0],recon.shape[0])
    angles = np.arange(np.min(omega), np.max(omega) + stepsize, stepsize)
    cos_a, sin_a = np.cos(np.radians(angles)), np.sin(np.radians(angles))
    center = recon.shape[0] // 2
    image = np.abs(recon)
    rotated = []

    # Rotate reconstruction in steps of stepsize dgrs.
    for i, (c,s) in enumerate(zip(cos_a, sin_a)):
        R = np.array([[c, s, -center * (c + s - 1)],
                        [-s, c, -center * (c - s - 1)],
                        [0, 0, 1]])
        rotated.append( warp(image, R, clip=False) )

    # Map to closest rotated reconstruction.
    cmsx = np.zeros((len(omega),))
    iy = np.round( (dty - ymin) / ystep ).astype(int)
    for i,(om,yindx) in enumerate(zip(omega,iy)):
        indx  = np.argmin( np.abs( angles-om ) )
        cmsx[i] = ( (np.sum( rotated[indx][:,yindx]*cms_weights  ) / np.sum( rotated[indx][:,yindx] )) - center ) * ystep

    return cmsx


import numpy as np
import matplotlib.pyplot as plt

from shapely.geometry import Polygon as shapelyPolygon
from scipy.optimize import minimize
from scipy.sparse import csr_matrix
from scipy.optimize import LinearConstraint

def _get_A_matrix_row( centroids, polymesh, n, N, entry, exit, nhat, beam_width ):
    
    zhat = np.array([0, 0, 1])
    that = np.cross( nhat, zhat )
    a = entry[0:3]  - nhat*len(polymesh)*beam_width + that*(beam_width/2.)
    b = entry[0:3]  - nhat*len(polymesh)*beam_width - that*(beam_width/2.)
    c = entry[0:3]  + nhat*len(polymesh)*beam_width - that*(beam_width/2.)
    d = entry[0:3]  + nhat*len(polymesh)*beam_width + that*(beam_width/2.)
    polybeam = shapelyPolygon([(a[0], a[1]), (b[0], b[1]), (c[0], c[1]), (d[0], d[1])])
    
    ts    = nhat[0:2].dot( (centroids - entry[0:2]).T )
    nn = nhat[0:2].reshape(2,1)
    en = entry[0:2].reshape(2,)
    dist = np.linalg.norm( en + (ts*nn).T - centroids, axis=1 )
    mask = dist < beam_width*(1+np.sqrt(2))/2.
    areas = _compute_intersection_areas( polymesh, polybeam, mask, plot=False )

    if np.sum(areas)==0: 
        return None
    else:
        return np.repeat(areas,6)*_directional_weights(n, N)/np.sum(areas)

def _compute_intersection_areas( polymesh, polybeam, mask, plot=False ):

    intersection_areas = np.zeros(len(polymesh))

    for i,poly_elm in enumerate(polymesh):

        if mask[i]:
            intersection_areas[i] = polybeam.intersection(poly_elm).area
        
        if plot:
            print("mask[i]",mask[i])
            plt.plot(polybeam.exterior.xy[0],polybeam.exterior.xy[1])
            for m in polymesh:
                plt.plot(m.exterior.xy[0],m.exterior.xy[1],"k")
            plt.plot(poly_elm.exterior.xy[0],poly_elm.exterior.xy[1],"r")
            plt.title(intersection_areas[i])
            plt.show()

    return intersection_areas

def _directional_weights( n, N ):
    """
    Compute the directional weights of a particular w:y scan beam
    (to be used to form a row in the A matrix)
    """
    return np.array( [ n[0]*n[0], n[1]*n[1], n[2]*n[2], 2*n[1]*n[2], 2*n[0]*n[2], 2*n[0]*n[1] ]*N )


def _calc_A_matrix( mesh, directions, entry, exit, nhat, beam_width, verbose  ):
    M = directions.shape[1] # number of measurements
    N = mesh.shape[0] # number of elements

    A = np.zeros((M,6*N))
    polymesh = [shapelyPolygon([(e[0], e[1]), (e[2], e[3]), (e[4], e[5]), (e[6], e[7])]) for e in mesh]
    centroids = np.array( [ np.mean(mesh[:,0::2],axis=1), np.mean(mesh[:,1::2],axis=1)] ).T

    bad_equations = []
    count=0
    if verbose:
        print("Building projection matrix...")
    for k in range(M):
        row = _get_A_matrix_row( centroids, polymesh, directions[:,k], N, entry[:,k], exit[:,k], nhat[:,k], beam_width )
        if k%300==0:
            if verbose:
                print("Getting row number: "+str(k) +" of "+str(M))
        if row is None:
            bad_equations.append(k)
            count+=1
        else:
            A[k,:] = row
    if verbose:
        print("")
        print("Rank of A: "+str(np.linalg.matrix_rank(A)) )
        print("Total number of eqs: "+str( M) )
        print("Topology cutoff lead to "+str( 100.*count/float(M) )+" percent eqs to be unusuable")
        print("")
    return A, bad_equations

def _constraints(mesh, low_bound, high_bound):
    """
    Limit the difference in strain between to neighbouring elements
    A neighbour pair is defined as two elements sharing at least one node
    """
    c = []
    incl = []

    data = []
    row = []
    col = []
    curr_row = 0
    for i,elm in enumerate(mesh):
        indx = _find_index_of_neighbors(mesh, elm)
        for j in indx:
            if [i,j] in incl or [j,i] in incl: continue
            
            for k in range(6):

                row.append( curr_row )
                row.append( curr_row )
                curr_row+=1
                data.append(1.)
                col.append( (i*6)+k)
                data.append(-1.)
                col.append((j*6)+k)

            incl.append([i,j])
    r,c = curr_row, mesh.shape[0]*6

    c = csr_matrix( (data, (row, col)), shape=(r, c) )
    lb = np.ones(c.shape[0])*low_bound
    ub = np.ones(c.shape[0])*high_bound

    return lb, c, ub


def _find_index_of_neighbors(mesh, element):
    elm_side = np.max(element[0::2]) - np.min(element[0::2])
    index_neighbors = []
    for i,elm in enumerate(mesh):
        if sum(elm==element)==len(elm): continue
        is_neighbour = False
        for x,y in zip(elm[0::2],elm[1::2]):
            for x_e,y_e in zip(element[0::2],element[1::2]):
                if abs(x-x_e)<elm_side/10. and abs(y-y_e)<elm_side/10.:
                    is_neighbour = True
        if is_neighbour: index_neighbors.append(i)
    return np.array( index_neighbors )


def trust_constr_solve( mesh, 
                        directions, 
                        strains , 
                        entry, 
                        exit,
                        nhat,
                        weights, 
                        beam_width, 
                        grad_constraint, 
                        maxiter,
                        verbose=True ):
    """Compute a voxelated strain-tensor field weighted least squares fit to a series of line integral strain measures.
    
    Assigns strain tensors over a 2d input mesh on a per element basis.

    Args:
        mesh (:obj:`numpy array`): Coordinate of mesh element nodes, ```shape=(N,4)``` for quads.
        directions (:obj:`numpy array`): unit vectors along which :obj:`strains` apply.
        strains (:obj:`numpy array`): Average strains along lines.
        entry (:obj:`numpy array`): Xray grian entry points per line.
        exit (:obj:`numpy array`): Xray grian exit points per line.
        nhat (:obj:`numpy array`): Xray direction per integral in sample coordinate system.
        weights (:obj:`numpy array`): Per measurement weights, higher weight gives more impact of equation.
        beam_width (:obj:`float`): Width of beam in units of microns.
        grad_constraint (:obj:`float`): In x-y plane smoothing constraint of strain reconstruction. Deviations between 
            two neighbouring pixels cannot exceed this value for any strain component.
        maxiter (:obj:`int`): Maximum number of WLSQ iterations to perform.
        verbose (:obj:`bool`): If to print progress. Defaults to True.

    Returns:
        (:obj:`list` of :obj:`numpy array`): List of strain tensor components each with ```shape=mesh.shape```. 
        The order of components is "XX","YY","ZZ","YZ","XZ","XY".

    """

    nelm = mesh.shape[0]

    A, bad_equations = _calc_A_matrix( mesh, directions, entry, exit, nhat, beam_width, verbose )

    A = np.delete(A, bad_equations, axis=0)
    strains = np.delete(strains, bad_equations, axis=0)
    weights = np.delete(weights, bad_equations, axis=0)

    if verbose:
        print("Computing _constraints matrix")
    lb,c,ub = _constraints(mesh, -grad_constraint, grad_constraint)

    linear_constraint = LinearConstraint(c, lb, ub, keep_feasible=True)
    x0 = np.zeros(6*nelm)

    def callback( xk, state ):
        if verbose:
            out="   {}      {}      {}"
            if state.nit==1: print(out.format("iteration","cost","max strain grad") )
            print( out.format( state.nit, np.round(state.fun,9), np.max(np.abs(c.dot(xk))) ) )
        return state.nit==maxiter

    W = np.diag( weights )
    WA = np.dot(W,A)
    m = strains
    Wm = np.dot(W,m)
    WATWA = np.dot( WA.T, WA )
    WATWm = np.dot( WA.T, Wm )
    
    def func( x ): return 0.5*np.linalg.norm( (np.dot( WA, x ) - Wm) )**2
    def jac( x ): return ( np.dot(WATWA,x) - WATWm )
    def hess( x ): return WATWA

    res = minimize(func, x0, method="trust-constr", jac=jac, hess=hess,\
                    callback=callback, tol=1e-8, \
                    constraints=[linear_constraint],\
                    options={"disp": verbose, "maxiter":maxiter})

    s_tilde = res.x

    # reformat, each row is strain for the element
    s_tilde = s_tilde.reshape(nelm,6)

    return [s_tilde[:,0],s_tilde[:,1],s_tilde[:,2],s_tilde[:,3],s_tilde[:,4],s_tilde[:,5]]


import numpy as np
from scipy.interpolate import griddata,interp2d
import matplotlib.pyplot as plt
from xfab import tools

'''Module for defining various analytical strain fields
'''

def cantilevered_beam( t, h, l, E, nu, Py, Iyy, Pz, Izz, tensor=True ):
    '''Defines strain in a cantilevered beam gicen some material parameters
    '''
    
    # strain field
    Fxx = lambda x, y, z: Py*(l-x)*y/E/Iyy + Pz*(l-x)*z/E/Izz
    Fyy = lambda x, y, z: -nu*Py*(l-x)*y/E/Iyy + -nu*Pz*(l-x)*z/E/Izz
    Fzz = lambda x, y, z: -nu*Py*(l-x)*y/E/Iyy + -nu*Pz*(l-x)*z/E/Izz
    Fxy = lambda x, y, z: -(1+nu)/E*Py*(h**2/4-y**2)/2/Iyy
    Fxz = lambda x, y, z: -(1+nu)/E*Pz*(t**2/4-z**2)/2/Izz
    Fyz = lambda x, y, z: 0*x + 0*y +0*z

    if tensor:
        # input vector x get tensor strain
        strain = lambda x: np.array( [
                        [Fxx(x[0],x[1],x[2]),Fxy(x[0],x[1],x[2]),Fxz(x[0],x[1],x[2])],
                        [Fxy(x[0],x[1],x[2]),Fyy(x[0],x[1],x[2]),Fyz(x[0],x[1],x[2])],
                        [Fxz(x[0],x[1],x[2]),Fyz(x[0],x[1],x[2]),Fzz(x[0],x[1],x[2])]
                        ] )
        return strain
    else:
        # input scalar x,y,z get scalars Fxx, Fyy, Fzz, Fxy, Fxz, Fyz
        return Fxx, Fyy, Fzz, Fxy, Fxz, Fyz

def maxwell_linear_strain(self, compliance, a, b, c, d, tx, ty, tz):
    '''Produce a callable function as fun(x,y,z) <-- strain tensor 
    at coordinate x,y,z, such that the strain field is in equlibrium 
    given a specific linear elastic compliance matrix using the
    Maxwell stress functions 
        PHI_ij = [A, 0, 0]
                 [0, B, 0]
                 [0, 0, C]
    and setting
        A = B = C = f(x,y,z)
        f(x,y,z = a*(x-cx)**3 + b*(y-cy)**3 + c*(z-cx)**3 + d*(x-cx)*(y-cy)*(z-cz)
    this will lead to the strain being linear in the spatial coordinates x,y,z.
    '''

    d2A_dydy = lambda x,y,z: b * ( 6*(y-ty) )
    d2A_dzdz = lambda x,y,z: c * ( 6*(z-tz) )
    d2A_dydz = lambda x,y,z: d * ( (x-tx) )

    d2B_dxdx = lambda x,y,z: a * ( 6*(x-tx) )
    d2B_dzdz = lambda x,y,z: c * ( 6*(z-tz) )
    d2B_dzdx = lambda x,y,z: d * ( (y-ty) )

    d2C_dxdx = lambda x,y,z: a * ( 6*(x-tx) )
    d2C_dydy = lambda x,y,z: b * ( 6*(y-ty) )
    d2C_dxdy = lambda x,y,z: d * ( (z-tz) )

    sigma_xx = lambda x,y,z:  d2B_dzdz(x,y,z) + d2C_dydy(x,y,z)
    sigma_yy = lambda x,y,z:  d2C_dxdx(x,y,z) + d2A_dzdz(x,y,z)
    sigma_zz = lambda x,y,z:  d2A_dydy(x,y,z) + d2B_dxdx(x,y,z)
    sigma_xy = lambda x,y,z: -d2C_dxdy(x,y,z)
    sigma_xz = lambda x,y,z: -d2B_dzdx(x,y,z)
    sigma_yz = lambda x,y,z: -d2A_dydz(x,y,z)

    sigma = lambda x,y,z: np.array([ sigma_xx(x,y,z) , sigma_yy(x,y,z) , sigma_zz(x,y,z) , \
                                     sigma_xy(x,y,z) , sigma_xz(x,y,z) , sigma_yz(x,y,z)  ])

    def strain(x,y,z):
        eps_vec = compliance.dot( sigma(x,y,z) )
        eps = np.array([[ eps_vec[0]   ,  eps_vec[3]/2 , eps_vec[4]/2  ],
                        [ eps_vec[3]/2 ,  eps_vec[1]   , eps_vec[5]/2  ],
                        [ eps_vec[4]/2 ,  eps_vec[5]/2 , eps_vec[2]    ]])
        return eps

    return strain

def field_from_simulation( paths, zpos ):
    '''Read the strain field from a list of simulation files, and return a 
    callable function that returns interpolated strain at specified X,Y,Z-
    grid coordinates. If only X and Y is passed, the read data in paths is 
    assumed to be 2D.
    '''
    coordinates, strains, euler_angles = [],[],[]
    
    for z,path in zip(zpos, paths):
        voxel_id = 0
        U = np.zeros((3,3))
        with open(path) as f:
            inp = f.readlines()
            for i,line in enumerate(inp):
                if 'pos_voxels_'+str(voxel_id) in line:
                    coordinates.append( [ float(line.split()[1]), float(line.split()[2]), z/float(1e3) ] )
                if 'U_voxels_'+str(voxel_id) in line:
                    U[0,0], U[0,1], U[0,2], U[1,0], U[1,1], U[1,2], U[2,0], U[2,1], U[2,2] = [float(number) for number in line.split()[1:]]
                if 'eps_voxels_'+str(voxel_id) in line:
                    eps11, eps12, eps13, eps22, eps23, eps33 = [float(number) for number in line.split()[1:]]
                    strain_tensor = np.array([[eps11,eps12,eps13],
                                              [eps12,eps22,eps23],
                                              [eps13,eps23,eps33]])
                    strain_tensor = ( U.dot( strain_tensor ) ).dot( U.T ) # to sample system
                    strains.append( [ strain_tensor[0,0], strain_tensor[1,1], strain_tensor[2,2], strain_tensor[0,1], strain_tensor[0,2],  strain_tensor[1,2] ] )
                    euler1, euler2, euler3 =  tools.u_to_euler( U )
                    euler_angles.append([euler1, euler2, euler3])
                    voxel_id+=1
    coordinates, strains = np.array(coordinates)*1e3, np.array(strains) # mm scaled to microns
    euler_angles = np.degrees( np.array( euler_angles ) )

    def strain_function( X, Y, Z=None ):
        
        if Z is None:
            points = coordinates[:,0:2]
            xi = np.array([ X.flatten(), Y.flatten() ]).T
        else:
            points = coordinates
            xi = np.array([ X.flatten(), Y.flatten(), Z.flatten() ]).T

        interp_strain = []
        for i in range(6):
            s = griddata(points, strains[:,i], xi, method='nearest')
            if Z is None:
                s = s.reshape(X.shape[0], X.shape[1])
            else:
                s = s.reshape(X.shape[0], X.shape[1], X.shape[2])
            interp_strain.append( s )
        return interp_strain

    def euler_function( X, Y, Z=None ):
        
        if Z is None:
            points = coordinates[:,0:2]
            xi = np.array([ X.flatten(), Y.flatten() ]).T
        else:
            points = coordinates
            xi = np.array([ X.flatten(), Y.flatten(), Z.flatten() ]).T

        interp_euler = []
        for i in range(3):
            e = griddata(points, euler_angles[:,i], xi, method='nearest')
            if Z is None:
                e = e.reshape(X.shape[0], X.shape[1])
            else:
                e = e.reshape(X.shape[0], X.shape[1], X.shape[2])
            interp_euler.append( e )
        return interp_euler

    return strain_function, euler_function, coordinates


def discrete_to_analytic( strains, xvec, yvec ):
    ''' Get an analytical function for evaluating strain at points in a plane. 
    strain in format:
        strain=[ XX , YY, ZZ , XY , XZ , YZ]
    and X,Y,Z are meshgrids
    '''
    eps_xx_func = interp2d(xvec, yvec, strains[0],  kind='linear')
    eps_yy_func = interp2d(xvec, yvec, strains[1],  kind='linear')
    eps_zz_func = interp2d(xvec, yvec, strains[2],  kind='linear')
    eps_xy_func = interp2d(xvec, yvec, strains[3],  kind='linear')
    eps_xz_func = interp2d(xvec, yvec, strains[4],  kind='linear')
    eps_yz_func = interp2d(xvec, yvec, strains[5],  kind='linear')
    def interp_strain( x ):
        eps_xx = eps_xx_func(x[0],x[1])[0]
        eps_yy = eps_yy_func(x[0],x[1])[0]
        eps_zz = eps_zz_func(x[0],x[1])[0]
        eps_xy = eps_xy_func(x[0],x[1])[0]
        eps_xz = eps_xz_func(x[0],x[1])[0]
        eps_yz = eps_yz_func(x[0],x[1])[0]
        eps = np.array([[eps_xx, eps_xy, eps_xz],
                        [eps_xy, eps_yy, eps_yz],
                        [eps_xz, eps_yz, eps_zz]])
        return eps
    return interp_strain







import setuptools

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="s3dxrd",
    version="0.0.1",
    author="Axel Henningsson",
    author_email="nilsaxelhenningsson@gmail.com",
    description="Tools for intragranular strain estimation with s3dxrd data.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://bitbucket.org/Axiomel/gp-xrd/src/multigrain/",
    project_urls={
        "Documentation": "https://bitbucket.org/Axiomel/gp-xrd/src/multigrain/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
)


import numpy as np
import shapely

def rotate_compliance( U, H ):
    """Rotate a series of 6x6 "tensor matrices" according to a series of 3x3 rotation matrices.
    
    Args:
        U (:obj:`numpy array`): Orientation/rotation tensors, ```shape=(Nx3x3)``` 
        H (:obj:`numpy array`): Compliance matrices, ```shape=(Nx6x6)```

    Returns:
        (:obj:`numpy array`) of ```shape=(Nx6x6)``` rotated compliance matrices.

    """
    R = _get_rotation_matrix( U )
    H_rot = np.zeros( H.shape )
    for i in range(H.shape[0]):
        H_rot[i,:,:] = (R[i,:,:].dot(H[i,:,:])).dot( np.linalg.inv(R[i,:,:]) )
    return H_rot

def _get_rotation_matrix( U ):
    """Return the rotation matrix, R, that corresponds to rotating a symmetric 3x3 tensor eps by U as U*eps*U.T.
    
    R*eps_bar contains the same values in vector format where eps_bar is 6x1 stack of eps uniqie values as

    Args:
        U (:obj:`numpy array`): Orientation/rotation tensors, ```shape=(Nx3x3)``` 

    Returns:
        (:obj:`numpy array`) of ```shape=(Nx6x6)``` rotation matrix.

    """
    # Format verified by sympy (see rotations.py in docs)
    R = np.array([[U[:,0,0]**2, U[:,0,1]**2, U[:,0,2]**2, 2*U[:,0,0]*U[:,0,1], 2*U[:,0,0]*U[:,0,2], 2*U[:,0,1]*U[:,0,2]], 
        [U[:,1,0]**2, U[:,1,1]**2, U[:,1,2]**2, 2*U[:,1,0]*U[:,1,1], 2*U[:,1,0]*U[:,1,2], 2*U[:,1,1]*U[:,1,2]], 
        [U[:,2,0]**2, U[:,2,1]**2, U[:,2,2]**2, 2*U[:,2,0]*U[:,2,1], 2*U[:,2,0]*U[:,2,2], 2*U[:,2,1]*U[:,2,2]], 
        [U[:,0,0]*U[:,1,0], U[:,0,1]*U[:,1,1], U[:,0,2]*U[:,1,2], U[:,0,0]*U[:,1,1] + U[:,0,1]*U[:,1,0], U[:,0,0]*U[:,1,2] + U[:,0,2]*U[:,1,0], U[:,0,1]*U[:,1,2] + U[:,0,2]*U[:,1,1]], 
        [U[:,0,0]*U[:,2,0], U[:,0,1]*U[:,2,1], U[:,0,2]*U[:,2,2], U[:,0,0]*U[:,2,1] + U[:,0,1]*U[:,2,0], U[:,0,0]*U[:,2,2] + U[:,0,2]*U[:,2,0], U[:,0,1]*U[:,2,2] + U[:,0,2]*U[:,2,1]], 
        [U[:,1,0]*U[:,2,0], U[:,1,1]*U[:,2,1], U[:,1,2]*U[:,2,2], U[:,1,0]*U[:,2,1] + U[:,1,1]*U[:,2,0], U[:,1,0]*U[:,2,2] + U[:,1,2]*U[:,2,0], U[:,1,1]*U[:,2,2] + U[:,1,2]*U[:,2,1]]])
    R = np.moveaxis(R, 0, 2)
    R = np.moveaxis(R, 0, 2)
    return R

def build_per_point_compliance_and_mask( compliance, Xgrid, Ygrid, Zgrid, polygons, polygon_orientations, zpos, ystep ):
    """Given a material compliance and orientation field, build a per point complance field and density mask.

    This function iterates over ```polygons``` to check if the points in ```Xgrid```, ```Ygrid```, ```Zgrid```
    are contained by any polygon. If so, the compliance of the given polygon is assigned to the point, as specified
    by polygon_orientations and compliance.

    Args:
        compliance (:obj:`numpy array`): Compliance matrix, ``shape=(6,6)``.
        Xgrid, Ygrid, Zgrid (:obj:`numpy array`): x,y,z coordinate arrays, ``shape=(m,n,l)``.
        polygons (:obj:`dict`): See s3dxrd.measurements.Id11.peaks_to_vectors()
        polygon_orientations (:obj:`dict`): See s3dxrd.measurements.Id11.peaks_to_vectors()
        zpos (:obj:`list`): z-coordinates of polygons.
        ystep (:obj:`float`): Distance in microns between gridpoints.

    Returns:
        (:obj:`tuple`): tuple with values:

            **Hpoint** (:obj:`numpy array`): Per gridpoint compliance of ``shape=(m,n,l,6,6)``. Zero outside of mask.
  
            **mask** (:obj:`numpy array`):  Density mask of ``shape=(m,n,l)``. np.nan is assigned to empty points.
  
    """
    Hpoint = np.zeros(  (Xgrid.shape[0],Xgrid.shape[1],Xgrid.shape[2], 6, 6) )
    mask = np.full( Xgrid.shape, np.nan )
    grainH  = {}
    for key1 in polygon_orientations:
        grainH[key1] = {}
        for key2 in polygon_orientations[key1]:
            orient = polygon_orientations[key1][key2]
            grainH[key1][key2] = rotate_compliance( orient.reshape(1,3,3), compliance.reshape(1,6,6 ) )
    for k in range(Xgrid.shape[2]):
        polygon_z_slice_indx = str(np.argmin( np.abs( zpos - Zgrid[0,0,k] ) ))
        for i in range(Xgrid.shape[0]):
            for j in range(Xgrid.shape[1]):
                p = shapely.geometry.Point(Xgrid[i,j,k],Ygrid[i,j,k])
                for key in polygons:
                    if polygon_z_slice_indx in polygons[key]:
                        if polygons[key][polygon_z_slice_indx].buffer( ystep/4. ).contains( p ):
                            Hpoint[i,j,k,:,:] = grainH[key][polygon_z_slice_indx]
                            if np.min(np.abs( zpos - Zgrid[i,j,k] ))<=ystep/2.:
                                mask[i,j,k] = 1
                            break
    Hpoint = Hpoint.reshape(Hpoint.shape[0]*Hpoint.shape[1]*Hpoint.shape[2],6,6)
    return Hpoint, mask